"""Service provider implementations."""

import hashlib
from typing import Optional, Sequence, Union
from weakref import ReferenceType

from ..utils.classloader import DeferLoad
from ..utils.stats import Collector
from .base import BaseInjector, BaseProvider, BaseSettings, InjectionError


class InstanceProvider(BaseProvider):
    """Provider for a previously-created instance."""

    def __init__(self, instance):
        """Initialize the instance provider."""
        if instance is None:
            raise ValueError("Class instance binding must be non-empty")
        self._instance = instance

    def provide(self, config: BaseSettings, injector: BaseInjector):
        """Provide the object instance given a config and injector."""
        inst = self._instance
        if isinstance(inst, ReferenceType):
            inst = inst()
            if inst is None:
                raise InjectionError("Weakref instance expired")
        return inst


class ClassProvider(BaseProvider):
    """Provider for a particular class."""

    class Inject:
        """A class for passing injected arguments to the constructor."""

        def __init__(self, base_cls: type):
            """Initialize the injected argument."""
            self.base_class = base_cls

    def __init__(
        self,
        instance_cls: Union[str, type],
        *ctor_args,
        init_method: Optional[str] = None,
        **ctor_kwargs,
    ):
        """Initialize the class provider."""
        self._ctor_args = ctor_args
        self._ctor_kwargs = ctor_kwargs
        self._init_method = init_method
        if isinstance(instance_cls, str):
            instance_cls = DeferLoad(instance_cls)
        self._instance_cls = instance_cls

    def provide(self, config: BaseSettings, injector: BaseInjector):
        """Provide the object instance given a config and injector."""
        args = []
        for arg in self._ctor_args:
            if isinstance(arg, ClassProvider.Inject):
                arg = injector.inject(arg.base_class)
            elif isinstance(arg, ReferenceType):
                arg = arg()
                if arg is None:
                    raise InjectionError("Weakref instance expired")
            args.append(arg)
        kwargs = {}
        for arg_name, arg in self._ctor_kwargs.items():
            if isinstance(arg, ClassProvider.Inject):
                arg = injector.inject(arg.base_class)
            elif isinstance(arg, ReferenceType):
                arg = arg()
                if arg is None:
                    raise InjectionError("Weakref instance expired")
            kwargs[arg_name] = arg
        instance = (self._instance_cls)(*args, **kwargs)
        if self._init_method:
            getattr(instance, self._init_method)()
        return instance


class CachedProvider(BaseProvider):
    """Cache the result of another provider."""

    def __init__(self, provider: BaseProvider, unique_settings_keys: tuple = ()):
        """Initialize the cached provider instance."""
        if not provider:
            raise ValueError("Cache provider input must not be empty.")
        self._instances = {}
        self._provider = provider
        self._unique_settings_keys = unique_settings_keys

    def provide(self, config: BaseSettings, injector: BaseInjector):
        """Provide the object instance given a config and injector.

        Instances are cached keyed on a SHA256 digest of the relevant subset
        of settings.
        """
        # MTODO: how to handle changes in the config?
        instance_vals = {key: config.get(key) for key in self._unique_settings_keys}
        instance_key = hashlib.sha256(str(instance_vals).encode()).hexdigest()
        if not self._instances.get(instance_key):
            self._instances[instance_key] = self._provider.provide(config, injector)

        return self._instances[instance_key]


class StatsProvider(BaseProvider):
    """Add statistics to the results of another provider."""

    def __init__(
        self,
        provider: BaseProvider,
        methods: Sequence[str],
        *,
        ignore_missing: bool = True,
    ):
        """Initialize the statistics provider instance."""
        if not provider:
            raise ValueError("Stats provider input must not be empty.")
        self._provider = provider
        self._methods = methods
        self._ignore_missing = ignore_missing

    def provide(self, config: BaseSettings, injector: BaseInjector):
        """Provide the object instance given a config and injector."""
        instance = self._provider.provide(config, injector)
        if self._methods:
            collector: Collector = injector.inject_or(Collector)
            if collector:
                collector.wrap(
                    instance, self._methods, ignore_missing=self._ignore_missing
                )
        return instance



import asyncio
import json

import pytest
from aiohttp.test_utils import AioHTTPTestCase, unused_port

from ....config.injection_context import InjectionContext
from ....tests import mock
from ....utils.testing import create_test_profile
from ...outbound.message import OutboundMessage
from ...wire_format import JsonWireFormat
from .. import ws as test_module
from ..message import InboundMessage
from ..session import InboundSession
from ..ws import WsTransport


class TestWsTransport(AioHTTPTestCase):
    async def asyncSetUp(self):
        self.message_results = []
        self.port = unused_port()
        self.session = None
        self.profile = await create_test_profile()
        self.transport = WsTransport(
            "0.0.0.0", self.port, self.create_session, root_profile=self.profile
        )
        self.transport.wire_format = JsonWireFormat()
        self.result_event = None
        await super().asyncSetUp()

    def get_profile(self):
        return self.profile

    def create_session(
        self,
        transport_type,
        *,
        client_info,
        wire_format,
        can_respond: bool = False,
        **kwargs,
    ):
        if not self.session:
            session = InboundSession(
                profile=self.get_profile(),
                can_respond=can_respond,
                inbound_handler=self.receive_message,
                session_id=None,
                wire_format=wire_format,
                client_info=client_info,
                transport_type=transport_type,
            )
            self.session = session
        result = asyncio.Future()
        result.set_result(self.session)
        return result

    def get_application(self):
        return self.transport.make_application()

    def receive_message(
        self,
        context: InjectionContext,
        message: InboundMessage,
        can_respond: bool = False,
    ):
        self.message_results.append((message.payload, message.receipt, can_respond))
        if self.result_event:
            self.result_event.set()

    async def test_start_x(self):
        with mock.patch.object(test_module.web, "TCPSite", mock.MagicMock()) as mock_site:
            mock_site.return_value = mock.MagicMock(
                start=mock.CoroutineMock(side_effect=OSError())
            )
            with pytest.raises(test_module.InboundTransportSetupError):
                await self.transport.start()

    async def test_message_and_response(self):
        await self.transport.start()

        test_message = {"test": "message"}
        test_response = {"response": "ok"}

        async with self.client.ws_connect("/") as ws:
            self.result_event = asyncio.Event()
            await ws.send_json(test_message)
            await asyncio.wait_for(self.result_event.wait(), timeout=0.1)

            assert self.session is not None
            assert len(self.message_results) == 1
            received, _, can_respond = self.message_results[0]
            assert received == test_message
            assert can_respond

            response = OutboundMessage(
                payload=None, enc_payload=json.dumps(test_response)
            )
            self.session.set_response(response)

            result = await asyncio.wait_for(ws.receive_json(), 1.0)
            assert result == {"response": "ok"}

            # Check the Server header is removed
            assert "Server" not in ws._response.headers

        await self.transport.stop()



from unittest import IsolatedAsyncioTestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ATTACHMENT_FORMAT, CRED_20_REQUEST
from ..cred_format import V20CredFormat
from ..cred_request import V20CredRequest


class TestV20CredRequest(IsolatedAsyncioTestCase):
    """Credential request tests"""

    indy_cred_req = {
        "nonce": "1017762706737386703693758",
        "prover_did": "GMm4vMw8LLrLJjp81kRRLp",
        "cred_def_id": "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag",
        "blinded_ms": {
            "u": "83907504917598709544715660183444547664806528194879236493704185267249518487609477830252206438464922282419526404954032744426656836343614241707982523911337758117991524606767981934822739259321023980818911648706424625657217291525111737996606024710795596961607334766957629765398381678917329471919374676824400143394472619220909211861028497009707890651887260349590274729523062264675018736459760546731362496666872299645586181905130659944070279943157241097916683504866583173110187429797028853314290183583689656212022982000994142291014801654456172923356395840313420880588404326139944888917762604275764474396403919497783080752861",
            "ur": "1 2422A7A25A9AB730F3399C77C28E1F6E02BB94A2C07D245B28DC4EE33E33DE49 1 1EF3FBD36FBA7510BDA79386508C0A84A33DF4171107C22895ACAE4FA4499F02 2 095E45DDF417D05FB10933FFC63D474548B7FFFF7888802F07FFFFFF7D07A8A8",
            "hidden_attributes": ["master_secret"],
            "committed_attributes": {},
        },
        "blinded_ms_correctness_proof": {
            "c": "77782990462020711078900471139684606615516190979556618670020830699801678914552",
            "v_dash_cap": "1966215015532422356590954855080129096516569112935438312989092847889400013191094311374123910677667707922694722167856889267996544544770134106600289624974901761453909338477897555013062690166110508298265469948048257876547569520215226798025984795668101468265482570744011744194025718081101032551943108999422057478928838218205736972438022128376728526831967897105301274481454020377656694232901381674223529320224276009919370080174601226836784570762698964476355045131401700464714725647784278935633253472872446202741297992383148244277451017022036452203286302631768247417186601621329239603862883753434562838622266122331169627284313213964584034951472090601638790603966977114416216909593408778336960753110805965734708636782885161632",
            "m_caps": {
                "master_secret": "1932933391026030434402535597188163725022560167138754201841873794167337347489231254032687761158191503499965986291267527620598858412377279828812688105949083285487853357240244045442"
            },
            "r_caps": {},
        },
    }

    ld_proof_cred_req = {
        "credential": {
            "@context": [
                "https://www.w3.org/2018/credentials/v1",
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "credentialSubject": {"test": "key"},
            "issuanceDate": "2021-04-12",
            "issuer": "did:sov:something",
        },
        "options": {"proofType": "Ed25519Signature2018"},
    }

    CRED_REQUEST = V20CredRequest(
        comment="Test",
        formats=[
            V20CredFormat(
                attach_id="indy",
                format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][V20CredFormat.Format.INDY.api],
            )
        ],
        requests_attach=[
            AttachDecorator.data_base64(
                ident="indy",
                mapping=indy_cred_req,
            )
        ],
    )

    async def test_init_type(self):
        """Test initializer and type."""
        assert (
            TestV20CredRequest.CRED_REQUEST.requests_attach[0].content
            == TestV20CredRequest.indy_cred_req
        )
        assert (
            TestV20CredRequest.CRED_REQUEST.attachment()
            == TestV20CredRequest.indy_cred_req
        )
        assert TestV20CredRequest.CRED_REQUEST._type == DIDCommPrefix.qualify_current(
            CRED_20_REQUEST
        )

    async def test_attachment_no_target_format(self):
        """Test attachment behaviour for only unknown formats."""

        x_cred_req = V20CredRequest(
            comment="Test",
            formats=[V20CredFormat(attach_id="not_indy", format_="not_indy")],
            requests_attach=[
                AttachDecorator.data_base64(
                    ident="not_indy",
                    mapping=TestV20CredRequest.indy_cred_req,
                )
            ],
        )
        assert x_cred_req.attachment() is None

    async def test_serde(self):
        """Test de/serialization."""
        obj = TestV20CredRequest.CRED_REQUEST.serialize()

        cred_request = V20CredRequest.deserialize(obj)
        assert type(cred_request) is V20CredRequest

        obj["requests~attach"][0]["data"]["base64"] = "eyJub3QiOiAiaW5keSJ9"
        with self.assertRaises(BaseModelError):
            V20CredRequest.deserialize(obj)

        obj["requests~attach"][0]["@id"] = "xxx"
        with self.assertRaises(BaseModelError):
            V20CredRequest.deserialize(obj)

        obj["requests~attach"].append(  # more attachments than formats
            {
                "@id": "def",
                "mime-type": "application/json",
                "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
            }
        )
        with self.assertRaises(BaseModelError):
            V20CredRequest.deserialize(obj)

        cred_request.formats.append(  # unknown format: no validation
            V20CredFormat(
                attach_id="not_indy",
                format_="not_indy",
            )
        )
        obj = cred_request.serialize()
        obj["requests~attach"].append(
            {
                "@id": "not_indy",
                "mime-type": "application/json",
                "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
            }
        )
        V20CredRequest.deserialize(obj)


class TestV20CredRequestSchema(IsolatedAsyncioTestCase):
    """Test credential request schema"""

    async def test_make_model(self):
        """Test making model."""
        cred_request = V20CredRequest(
            comment="Test",
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                ),
                V20CredFormat(
                    attach_id="ld-proof-json",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                ),
            ],
            requests_attach=[
                AttachDecorator.data_base64(
                    ident="indy", mapping=TestV20CredRequest.indy_cred_req
                ),
                AttachDecorator.data_json(
                    ident="ld-proof-json", mapping=TestV20CredRequest.ld_proof_cred_req
                ),
            ],
        )

        assert (
            cred_request.attachment(V20CredFormat.Format.INDY)
            == TestV20CredRequest.indy_cred_req
        )
        assert (
            cred_request.attachment(V20CredFormat.Format.LD_PROOF)
            == TestV20CredRequest.ld_proof_cred_req
        )
        data = cred_request.serialize()
        model_instance = V20CredRequest.deserialize(data)
        assert isinstance(model_instance, V20CredRequest)

        cred_request = V20CredRequest(
            comment="Test",
            formats=[
                V20CredFormat(
                    attach_id="indy",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                ),
                V20CredFormat(
                    attach_id="ld-proof-links",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                ),
            ],
            requests_attach=[
                AttachDecorator.data_base64(
                    ident="indy", mapping=TestV20CredRequest.indy_cred_req
                ),
                AttachDecorator.data_links(
                    ident="ld-proof-links",
                    links="http://10.20.30.40/cred-req.json",
                    sha256="00000000000000000000000000000000",
                ),
            ],
        )
        assert cred_request.attachment(V20CredFormat.Format.LD_PROOF) == (
            ["http://10.20.30.40/cred-req.json"],
            "00000000000000000000000000000000",
        )



"""Linked data proof specific credential exchange information with non-secrets storage."""

from typing import Any, Optional, Sequence

from marshmallow import EXCLUDE, fields

from ......core.profile import ProfileSession
from ......messaging.models.base_record import BaseRecord, BaseRecordSchema
from ......messaging.valid import UUID4_EXAMPLE
from .. import UNENCRYPTED_TAGS


class V20CredExRecordLDProof(BaseRecord):
    """Credential exchange linked data proof detail record."""

    class Meta:
        """V20CredExRecordLDProof metadata."""

        schema_class = "V20CredExRecordLDProofSchema"

    RECORD_ID_NAME = "cred_ex_ld_proof_id"
    RECORD_TYPE = "ld_proof_cred_ex_v20"
    TAG_NAMES = {"~cred_ex_id"} if UNENCRYPTED_TAGS else {"cred_ex_id"}
    RECORD_TOPIC = "issue_credential_v2_0_ld_proof"

    def __init__(
        self,
        cred_ex_ld_proof_id: Optional[str] = None,
        *,
        cred_ex_id: Optional[str] = None,
        cred_id_stored: Optional[str] = None,
        **kwargs,
    ):
        """Initialize LD Proof credential exchange record details."""
        super().__init__(cred_ex_ld_proof_id, **kwargs)

        self.cred_ex_id = cred_ex_id
        self.cred_id_stored = cred_id_stored

    @property
    def cred_ex_ld_proof_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this credential exchange."""
        return {prop: getattr(self, prop) for prop in ("cred_id_stored",)}

    @classmethod
    async def query_by_cred_ex_id(
        cls,
        session: ProfileSession,
        cred_ex_id: str,
    ) -> Sequence["V20CredExRecordLDProof"]:
        """Retrieve a credential exchange LDProof detail record by its cred ex id."""
        return await cls.query(
            session=session,
            tag_filter={"cred_ex_id": cred_ex_id},
        )

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class V20CredExRecordLDProofSchema(BaseRecordSchema):
    """Credential exchange linked data proof detail record detail schema."""

    class Meta:
        """Credential exchange linked data proof detail record schema metadata."""

        model_class = V20CredExRecordLDProof
        unknown = EXCLUDE

    cred_ex_ld_proof_id = fields.Str(
        required=False,
        metadata={"description": "Record identifier", "example": UUID4_EXAMPLE},
    )
    cred_ex_id = fields.Str(
        required=False,
        metadata={
            "description": "Corresponding v2.0 credential exchange record identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    cred_id_stored = fields.Str(
        required=False,
        metadata={
            "description": "Credential identifier stored in wallet",
            "example": UUID4_EXAMPLE,
        },
    )



"""Record for out of band invitations."""

from typing import Any, Mapping, Optional, Union

from marshmallow import fields

from .....messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from .....messaging.valid import UUID4_EXAMPLE
from ..messages.invitation import InvitationMessage, InvitationMessageSchema


class InvitationRecord(BaseExchangeRecord):
    """Represents an out of band invitation record."""

    class Meta:
        """InvitationRecord metadata."""

        schema_class = "InvitationRecordSchema"

    RECORD_TYPE = "oob_invitation"
    RECORD_ID_NAME = "invitation_id"
    RECORD_TOPIC = "oob_invitation"
    TAG_NAMES = {"invi_msg_id"}

    STATE_INITIAL = "initial"
    STATE_AWAIT_RESPONSE = "await_response"
    STATE_DONE = "done"

    def __init__(
        self,
        *,
        invitation_id: Optional[str] = None,
        state: Optional[str] = None,
        invi_msg_id: Optional[str] = None,
        invitation: Union[InvitationMessage, Mapping] = None,  # invitation message
        invitation_url: Optional[str] = None,
        oob_id: Optional[str] = None,
        public_did: Optional[str] = None,  # backward-compat: BaseRecord.from_storage()
        trace: bool = False,
        **kwargs,
    ):
        """Initialize a new InvitationRecord."""
        super().__init__(invitation_id, state, trace=trace, **kwargs)
        self._id = invitation_id
        self.state = state
        self.invi_msg_id = invi_msg_id
        self._invitation = InvitationMessage.serde(invitation)
        self.invitation_url = invitation_url
        self.oob_id = oob_id
        self.trace = trace

    @property
    def invitation_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def invitation(self) -> InvitationMessage:
        """Accessor; get deserialized view."""
        return None if self._invitation is None else self._invitation.de

    @invitation.setter
    def invitation(self, value):
        """Setter; store de/serialized views."""
        self._invitation = InvitationMessage.serde(value)

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this invitation."""
        return {
            **{
                prop: getattr(self, prop)
                for prop in ("invitation_url", "state", "trace", "oob_id")
            },
            **{
                prop: getattr(self, f"_{prop}").ser
                for prop in ("invitation",)
                if getattr(self, prop) is not None
            },
        }

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class InvitationRecordSchema(BaseExchangeSchema):
    """Schema to allow serialization/deserialization of invitation records."""

    class Meta:
        """InvitationRecordSchema metadata."""

        model_class = InvitationRecord

    invitation_id = fields.Str(
        required=False,
        metadata={
            "description": "Invitation record identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    state = fields.Str(
        required=False,
        metadata={
            "description": "Out of band message exchange state",
            "example": InvitationRecord.STATE_AWAIT_RESPONSE,
        },
    )
    invi_msg_id = fields.Str(
        required=False,
        metadata={
            "description": "Invitation message identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    oob_id = fields.Str(
        required=False,
        metadata={
            "description": "Out of band record identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    invitation = fields.Nested(
        InvitationMessageSchema(),
        required=False,
        metadata={"description": "Out of band invitation message"},
    )
    invitation_url = fields.Str(
        required=False,
        metadata={
            "description": "Invitation message URL",
            "example": (
                "https://example.com/endpoint?c_i=eyJAdHlwZSI6ICIuLi4iLCAiLi4uIjog"
                "Ii4uLiJ9XX0="
            ),
        },
    )



"""Http Transport classes and functions."""

import logging

from aiohttp import web

from ...messaging.error import MessageParseError
from ...utils.server import remove_unwanted_headers
from ..error import WireFormatParseError
from ..wire_format import DIDCOMM_V0_MIME_TYPE, DIDCOMM_V1_MIME_TYPE
from .base import BaseInboundTransport, InboundTransportSetupError

LOGGER = logging.getLogger(__name__)


class HttpTransport(BaseInboundTransport):
    """Http Transport class."""

    def __init__(self, host: str, port: int, create_session, **kwargs) -> None:
        """Initialize an inbound HTTP transport instance.

        Args:
            host: Host to listen on
            port: Port to listen on
            create_session: Method to create a new inbound session
            kwargs: Additional keyword arguments

        """
        super().__init__("http", create_session, **kwargs)
        self.host = host
        self.port = port
        self.site: web.BaseSite = None

    async def make_application(self) -> web.Application:
        """Construct the aiohttp application."""
        app_args = {}
        if self.max_message_size:
            app_args["client_max_size"] = self.max_message_size
        app = web.Application(**app_args)
        app.add_routes([web.get("/", self.invite_message_handler)])
        app.add_routes([web.post("/", self.inbound_message_handler)])
        app.on_response_prepare.append(remove_unwanted_headers)
        return app

    async def start(self) -> None:
        """Start this transport.

        Raises:
            InboundTransportSetupError: If there was an error starting the webserver

        """
        app = await self.make_application()
        runner = web.AppRunner(app)
        await runner.setup()
        self.site = web.TCPSite(runner, host=self.host, port=self.port)
        try:
            await self.site.start()
        except OSError:
            raise InboundTransportSetupError(
                "Unable to start webserver with host "
                + f"'{self.host}' and port '{self.port}'\n"
            )

    async def stop(self) -> None:
        """Stop this transport."""
        if self.site:
            await self.site.stop()
            self.site = None

    async def inbound_message_handler(self, request: web.BaseRequest):
        """Message handler for inbound messages.

        Args:
            request: aiohttp request object

        Returns:
            The web response

        """
        ctype = request.headers.get("content-type", "")
        if ctype.split(";", 1)[0].lower() == "application/json":
            body = await request.text()
        else:
            body = await request.read()

        client_info = {"host": request.host, "remote": request.remote}

        session = await self.create_session(
            accept_undelivered=True, can_respond=True, client_info=client_info
        )

        async with session:
            try:
                inbound = await session.receive(body)
            except (MessageParseError, WireFormatParseError):
                raise web.HTTPBadRequest()

            if inbound.receipt.direct_response_requested:
                # Wait for the message to be processed. Only send a response if a response
                # buffer is present.
                await inbound.wait_processing_complete()
                response = (
                    await session.wait_response() if session.response_buffer else None
                )

                # no more responses
                session.can_respond = False
                session.clear_response()

                if response:
                    if isinstance(response, bytes):
                        return web.Response(
                            body=response,
                            status=200,
                            headers={
                                "Content-Type": (
                                    DIDCOMM_V1_MIME_TYPE
                                    if session.profile.settings.get(
                                        "emit_new_didcomm_mime_type"
                                    )
                                    else DIDCOMM_V0_MIME_TYPE
                                )
                            },
                        )
                    else:
                        return web.Response(
                            text=response,
                            status=200,
                            headers={"Content-Type": "application/json"},
                        )
        return web.Response(status=200)

    async def invite_message_handler(self, request: web.BaseRequest):
        """Message handler for invites.

        Args:
            request: aiohttp request object

        Returns:
            The web response

        """
        if request.query.get("c_i"):
            return web.Response(
                text="You have received a connection invitation. To accept the "
                "invitation, paste it into your agent application."
            )
        else:
            return web.Response(status=200)



"""Test MultikeypManager."""

from unittest import IsolatedAsyncioTestCase

from acapy_agent.utils.testing import create_test_profile
from acapy_agent.wallet.key_type import KeyTypes
from acapy_agent.wallet.keys.manager import (
    MultikeyManager,
    multikey_to_verkey,
    verkey_to_multikey,
)


class TestKeyOperations(IsolatedAsyncioTestCase):
    seed = "00000000000000000000000000000000"

    ed25519_multikey = "z6MkgKA7yrw5kYSiDuQFcye4bMaJpcfHFry3Bx45pdWh3s8i"
    ed25519_verkey = "2ru5PcgeQzxF7QZYwQgDkG2K13PRqyigVw99zMYg8eML"
    ed25519_alg = "ed25519"

    p256_multikey = "zDnaeSd75MAwSRmem34MfZEzSMjQNcpWLmzkbF8Su49AuA9U2"
    p256_verkey = "demmi97mhJ7JQu31git4hQz8a1PD1dETJH9TVKaynNQv"
    p256_alg = "p256"

    bls12381g2_multikey = "zUC71fcKNvfU5d4NT3vurh8wdBqD2VSaVz7RdHmsfFBiYqfLDFkJTVK3m3hLb7yYDZq1C95HyssoX5BCr4ZatwP7jEh3UnwW7AMnx5fxdrhNkGVknbVY5QmjJ6S2CmtztCCffFL"
    bls12381g2_verkey = "mq4SKF1Ej1CA37G4gkSjKtUHnD8NU1ZVkuC4BPiuoxJXoenfkputxbjfS8dHhGHN3vfQwU1z9BdEuBqTjg3PuHAKgT3Q9XEJgRyZje4WGKMtRh9vzUbd8J23jbA7Je3oAe2"
    bls12381g2_alg = "bls12381g2"

    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())

    async def test_key_creation(self):
        async with self.profile.session() as session:
            for i, (alg, expected_multikey) in enumerate(
                [
                    (self.ed25519_alg, self.ed25519_multikey),
                    (self.p256_alg, self.p256_multikey),
                    (self.bls12381g2_alg, self.bls12381g2_multikey),
                ]
            ):
                manager = MultikeyManager(session=session)
                kid = f"did:web:example.com#key-0{i}"

                key_info = await manager.create(seed=self.seed, alg=alg)
                assert key_info["multikey"] == expected_multikey
                assert key_info["kid"] is None

                key_info = await manager.from_multikey(multikey=expected_multikey)
                assert key_info["multikey"] == expected_multikey
                assert key_info["kid"] == []

                key_info = await manager.update(multikey=expected_multikey, kid=kid)
                assert key_info["multikey"] == expected_multikey
                assert key_info["kid"] == kid

                key_info = await manager.from_kid(kid=kid)
                assert key_info["multikey"] == expected_multikey
                assert key_info["kid"] == kid

    async def test_key_id_binding(self):
        async with self.profile.session() as session:
            test_multikey = self.ed25519_multikey
            key_id_01 = "did:web:example.com#key-01"
            key_id_02 = "did:web:example.com#key-02"
            key_id_03 = "did:web:example.com#key-03"

            manager = MultikeyManager(session=session)

            await manager.create(self.seed, key_id_01, self.ed25519_alg)
            await manager.bind_key_id(test_multikey, key_id_02)
            await manager.bind_key_id(test_multikey, key_id_03)

            assert (await manager.from_kid(key_id_01))["multikey"] == test_multikey
            assert (await manager.from_kid(key_id_02))["multikey"] == test_multikey
            assert (await manager.from_kid(key_id_03))["multikey"] == test_multikey

            await manager.unbind_key_id(test_multikey, key_id_01)

            assert (await manager.from_kid(key_id_01)) is None
            assert (await manager.from_kid(key_id_02))["multikey"] == test_multikey
            assert (await manager.from_kid(key_id_03))["multikey"] == test_multikey

    async def test_key_transformations(self):
        for alg, multikey, verkey in [
            (self.ed25519_alg, self.ed25519_multikey, self.ed25519_verkey),
            (self.p256_alg, self.p256_multikey, self.p256_verkey),
            (self.bls12381g2_alg, self.bls12381g2_multikey, self.bls12381g2_verkey),
        ]:
            assert multikey_to_verkey(multikey) == verkey
            assert verkey_to_multikey(verkey, alg=alg) == multikey



from unittest import IsolatedAsyncioTestCase

from ...ledger.base import BaseLedger
from ...ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ...multitenant.base import BaseMultitenantManager
from ...multitenant.manager import MultitenantManager
from ...storage.error import StorageNotFoundError
from ...tests import mock
from ...utils.testing import create_test_profile
from ..error import (
    RevocationError,
    RevocationNotSupportedError,
    RevocationRegistryBadSizeError,
)
from ..indy import IndyRevocation
from ..models.issuer_rev_reg_record import DEFAULT_REGISTRY_SIZE, IssuerRevRegRecord
from ..models.revocation_registry import RevocationRegistry


class TestIndyRevocation(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = self.profile.context

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": True}}
        )
        self.ledger.get_revoc_reg_def = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_executor
        )
        self.revoc = IndyRevocation(self.profile)

        self.test_did = "sample-did"

    async def test_init_issuer_registry(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"

        result = await self.revoc.init_issuer_registry(CRED_DEF_ID)

        assert result.cred_def_id == CRED_DEF_ID
        assert result.issuer_did == self.test_did
        assert result.max_cred_num == DEFAULT_REGISTRY_SIZE
        assert result.revoc_def_type == IssuerRevRegRecord.REVOC_DEF_TYPE_CL
        assert result.tag is None

        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=(None, self.ledger)),
        ):
            result = await self.revoc.init_issuer_registry(CRED_DEF_ID)
        assert result.cred_def_id == CRED_DEF_ID
        assert result.issuer_did == self.test_did
        assert result.max_cred_num == DEFAULT_REGISTRY_SIZE
        assert result.revoc_def_type == IssuerRevRegRecord.REVOC_DEF_TYPE_CL
        assert result.tag is None

    async def test_init_issuer_registry_no_cred_def(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"

        self.profile.context.injector.clear_binding(BaseLedger)
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=None)
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)

        with self.assertRaises(RevocationNotSupportedError):
            await self.revoc.init_issuer_registry(CRED_DEF_ID)

    async def test_init_issuer_registry_bad_size(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"

        self.profile.context.injector.clear_binding(BaseLedger)
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": "..."}}
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)

        with self.assertRaises(RevocationRegistryBadSizeError):
            await self.revoc.init_issuer_registry(
                CRED_DEF_ID,
                max_cred_num=1,
            )

    async def test_get_active_issuer_rev_reg_record(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        rec = await self.revoc.init_issuer_registry(CRED_DEF_ID)
        rec.revoc_reg_id = "dummy"
        rec.state = IssuerRevRegRecord.STATE_ACTIVE

        async with self.profile.session() as session:
            await rec.save(session)

        result = await self.revoc.get_active_issuer_rev_reg_record(CRED_DEF_ID)
        assert rec == result

    async def test_get_active_issuer_rev_reg_record_none(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"
        with self.assertRaises(StorageNotFoundError):
            await self.revoc.get_active_issuer_rev_reg_record(CRED_DEF_ID)

    async def test_init_issuer_registry_no_revocation(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"

        self.profile.context.injector.clear_binding(BaseLedger)
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {}}
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)

        with self.assertRaises(RevocationNotSupportedError):
            await self.revoc.init_issuer_registry(CRED_DEF_ID)

    async def test_get_issuer_rev_reg_record(self):
        CRED_DEF_ID = f"{self.test_did}:3:CL:1234:default"

        rec = await self.revoc.init_issuer_registry(CRED_DEF_ID)
        rec.revoc_reg_id = "dummy"
        rec.generate_registry = mock.CoroutineMock()

        with mock.patch.object(
            IssuerRevRegRecord, "retrieve_by_revoc_reg_id", mock.CoroutineMock()
        ) as mock_retrieve_by_rr_id:
            mock_retrieve_by_rr_id.return_value = rec
            await rec.generate_registry(self.profile, None)

            result = await self.revoc.get_issuer_rev_reg_record(rec.revoc_reg_id)
            assert result.revoc_reg_id == "dummy"

    async def test_list_issuer_registries(self):
        CRED_DEF_ID = [f"{self.test_did}:3:CL:{i}:default" for i in (1234, 5678)]

        for cd_id in CRED_DEF_ID:
            await self.revoc.init_issuer_registry(cd_id)

        assert len(await self.revoc.list_issuer_registries()) == 2

    async def test_decommission_issuer_registries(self):
        CRED_DEF_ID = [f"{self.test_did}:3:CL:{i}:default" for i in (4321, 8765)]

        for cd_id in CRED_DEF_ID:
            await self.revoc.init_issuer_registry(cd_id)

        # 2 registries, both in init state (no listener to push into active)
        recs = await self.revoc.list_issuer_registries()
        assert len(recs) == 2

        init_list = list(filter(lambda r: r.state == IssuerRevRegRecord.STATE_INIT, recs))
        assert len(init_list) == 2

        # store the ids to verify they are decommissioned
        rev_reg_ids = [rec.revoc_reg_id for rec in recs if rec.revoc_reg_id]

        # need these to be not init so we can decommission
        async with self.profile.transaction() as txn:
            for rec in recs:
                registry = await IssuerRevRegRecord.retrieve_by_revoc_reg_id(
                    txn, rec.revoc_reg_id, for_update=True
                )
                await registry.set_state(
                    txn,
                    IssuerRevRegRecord.STATE_ACTIVE,
                )
            await txn.commit()

        # still 2 registries, but now active
        recs = await self.revoc.list_issuer_registries()
        assert len(recs) == 2
        active_list = list(
            filter(lambda r: r.state == IssuerRevRegRecord.STATE_ACTIVE, recs)
        )
        assert len(active_list) == 2

        #
        # decommission (active-> decommission, create replacement regs)
        #
        for cd_id in CRED_DEF_ID:
            rec = await self.revoc.decommission_registry(cd_id)

        # four entries, 2 new (init), 2 decommissioned
        recs = await self.revoc.list_issuer_registries()
        assert len(recs) == 4

        # previously active are decommissioned
        decomm_list = list(
            filter(lambda r: r.state == IssuerRevRegRecord.STATE_DECOMMISSIONED, recs)
        )
        assert len(decomm_list) == 2
        decomm_rev_reg_ids = [rec.revoc_reg_id for rec in decomm_list if rec.revoc_reg_id]

        # new ones replacing the decommissioned are in init state
        init_list = list(filter(lambda r: r.state == IssuerRevRegRecord.STATE_INIT, recs))
        assert len(init_list) == 2

        # check that the original rev reg ids are decommissioned
        rev_reg_ids.sort()
        decomm_rev_reg_ids.sort()
        assert rev_reg_ids == decomm_rev_reg_ids

    async def test_get_ledger_registry(self):
        with mock.patch.object(
            RevocationRegistry, "from_definition", mock.MagicMock()
        ) as mock_from_def:
            result = await self.revoc.get_ledger_registry("dummy")
            assert result == mock_from_def.return_value
            assert "dummy" in IndyRevocation.REV_REG_CACHE

            await self.revoc.get_ledger_registry("dummy")

        mock_from_def.assert_called_once_with(
            self.ledger.get_revoc_reg_def.return_value, True
        )

        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=(None, self.ledger)),
            ),
            mock.patch.object(
                RevocationRegistry, "from_definition", mock.MagicMock()
            ) as mock_from_def,
        ):
            result = await self.revoc.get_ledger_registry("dummy2")
            assert result == mock_from_def.return_value
            assert "dummy2" in IndyRevocation.REV_REG_CACHE

            await self.revoc.get_ledger_registry("dummy2")

        mock_from_def.assert_called_once_with(
            self.ledger.get_revoc_reg_def.return_value, True
        )

    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation.get_active_issuer_rev_reg_record",
        mock.CoroutineMock(
            return_value=mock.MagicMock(
                get_registry=mock.MagicMock(
                    return_value=mock.MagicMock(
                        get_or_fetch_local_tails_path=mock.CoroutineMock(
                            return_value="dummy"
                        )
                    )
                )
            )
        ),
    )
    async def test_get_or_create_active_registry_has_active_registry(self, *_):
        result = await self.revoc.get_or_create_active_registry("cred_def_id")
        assert isinstance(result, tuple)

    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation.get_active_issuer_rev_reg_record",
        mock.CoroutineMock(side_effect=StorageNotFoundError("No such record")),
    )
    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation.init_issuer_registry",
        mock.CoroutineMock(return_value=None),
    )
    @mock.patch.object(
        IssuerRevRegRecord,
        "query_by_cred_def_id",
        side_effect=[[], [IssuerRevRegRecord(max_cred_num=3)]],
    )
    async def test_get_or_create_active_registry_has_no_active_and_only_full_registies(
        self, *_
    ):
        result = await self.revoc.get_or_create_active_registry("cred_def_id")

        assert not result
        assert self.revoc.init_issuer_registry.call_args.kwargs["max_cred_num"] == 3

    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation.get_active_issuer_rev_reg_record",
        mock.CoroutineMock(side_effect=StorageNotFoundError("No such record")),
    )
    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation.init_issuer_registry",
        mock.CoroutineMock(return_value=None),
    )
    @mock.patch.object(IssuerRevRegRecord, "query_by_cred_def_id", side_effect=[[], []])
    async def test_get_or_create_active_registry_has_no_active_or_any_registry(self, *_):
        with self.assertRaises(RevocationError):
            await self.revoc.get_or_create_active_registry("cred_def_id")

    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation.get_active_issuer_rev_reg_record",
        mock.CoroutineMock(side_effect=StorageNotFoundError("No such record")),
    )
    @mock.patch(
        "acapy_agent.revocation.indy.IndyRevocation._set_registry_status",
        mock.CoroutineMock(return_value=None),
    )
    @mock.patch.object(
        IssuerRevRegRecord,
        "query_by_cred_def_id",
        side_effect=[
            [IssuerRevRegRecord(max_cred_num=3)],
            [
                IssuerRevRegRecord(
                    revoc_reg_id="test-rev-reg-id",
                    state=IssuerRevRegRecord.STATE_POSTED,
                )
            ],
        ],
    )
    async def test_get_or_create_active_registry_has_no_active_with_posted(self, *_):
        result = await self.revoc.get_or_create_active_registry("cred_def_id")

        assert not result
        assert (
            self.revoc._set_registry_status.call_args.kwargs["state"]
            == IssuerRevRegRecord.STATE_ACTIVE
        )



"""Presentation problem report message handler."""

from .....messaging.base_handler import BaseHandler
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from ..manager import V20PresManager
from ..messages.pres_problem_report import V20PresProblemReport


class V20PresProblemReportHandler(BaseHandler):
    """Message handler class for problem reports."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for problem reports.

        Args:
            context: request context
            responder: responder callback
        """
        self._logger.debug(
            "Present-proof v2.0 problem report handler called with context %s",
            context,
        )
        assert isinstance(context.message, V20PresProblemReport)

        pres_manager = V20PresManager(context.profile)
        try:
            await pres_manager.receive_problem_report(
                context.message,
                (
                    context.connection_record.connection_id
                    if context.connection_record is not None
                    else None
                ),
            )
        except StorageNotFoundError:
            self._logger.warning(
                "Record not found while processing present-proof v2.0 problem report"
            )
        except StorageError:
            self._logger.exception(
                "Storage error while processing present-proof v2.0 problem report"
            )



"""Message type identifiers for Connections."""

from ....messaging.util import get_proto_default_version
from ...didcomm_prefix import DIDCommPrefix
from ..definition import versions

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "25464a5c8f8a17b14edaa4310393df6094ace7b0/features/0023-did-exchange"
)
# Default Version
DEFAULT_VERSION = get_proto_default_version(versions, 1)
DIDEX_1_0 = "didexchange/1.0"
DIDEX_1_1 = "didexchange/1.1"
ARIES_PROTOCOL = f"didexchange/{DEFAULT_VERSION}"

# Message types
DIDX_REQUEST = f"{ARIES_PROTOCOL}/request"
DIDX_RESPONSE = f"{ARIES_PROTOCOL}/response"
DIDX_COMPLETE = f"{ARIES_PROTOCOL}/complete"
PROBLEM_REPORT = f"{ARIES_PROTOCOL}/problem_report"

PROTOCOL_PACKAGE = "acapy_agent.protocols.didexchange.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        DIDX_REQUEST: f"{PROTOCOL_PACKAGE}.messages.request.DIDXRequest",
        DIDX_RESPONSE: f"{PROTOCOL_PACKAGE}.messages.response.DIDXResponse",
        DIDX_COMPLETE: f"{PROTOCOL_PACKAGE}.messages.complete.DIDXComplete",
        PROBLEM_REPORT: (f"{PROTOCOL_PACKAGE}.messages.problem_report.DIDXProblemReport"),
    }
)



"""Rotate problem report handler."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..manager import DIDRotateManager
from ..messages.problem_report import RotateProblemReport


class ProblemReportHandler(BaseHandler):
    """Message handler class for rotate message."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle rotate problem report message.

        Args:
            context: request context
            responder: responder callback
        """
        self._logger.debug("ProblemReportHandler called with context %s", context)
        assert isinstance(context.message, RotateProblemReport)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        problem_report = context.message

        profile = context.profile
        did_rotate_mgr = DIDRotateManager(profile)

        await did_rotate_mgr.receive_problem_report(problem_report)



from unittest import TestCase, mock

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ACK, PROTOCOL_PACKAGE
from ..ack import RotateAck

THID = "test-thid"
PTHID = "test-pthid"


class TestRotateAck(TestCase):
    def test_init_type(self):
        """Test initializer."""

        obj = RotateAck()
        assert obj._type == DIDCommPrefix.qualify_current(ACK)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.ack.RotateAckSchema.load")
    def test_deserialize(self, mock_rotate_ack_schema_load):
        """Test deserialization."""

        obj = RotateAck()
        rotate_ack = RotateAck.deserialize(obj)
        mock_rotate_ack_schema_load.assert_called_once_with(obj)
        assert rotate_ack is mock_rotate_ack_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.ack.RotateAckSchema.dump")
    def test_serialize(self, mock_rotate_ack_schema_dump):
        """Test serialization."""

        obj = RotateAck()
        rotate_ack_dict = obj.serialize()
        mock_rotate_ack_schema_dump.assert_called_once_with(obj)
        assert rotate_ack_dict is mock_rotate_ack_schema_dump.return_value

    def test_serde(self):
        obj = {"~thread": {"thid": THID, "pthid": PTHID}, "status": "OK"}
        rotate_ack = RotateAck.deserialize(obj)
        assert rotate_ack._type == DIDCommPrefix.qualify_current(ACK)

        rotate_ack_dict = rotate_ack.serialize()
        assert rotate_ack_dict["~thread"] == obj["~thread"]

    def test_make_model(self):
        """Test making model."""

        obj = RotateAck()
        obj.assign_thread_id(THID, PTHID)
        rotate_ack_dict = obj.serialize()
        rotate_ack = RotateAck.deserialize(rotate_ack_dict)
        assert isinstance(rotate_ack, RotateAck)

    def test_make_model_fails_no_thread(self):
        """Test making model with no thread attachment."""

        obj = RotateAck()
        with self.assertRaises(BaseModelError):
            rotate_ack_dict = obj.serialize()



from unittest import TestCase

from ...protocols.out_of_band.v1_0.messages.invitation import InvitationMessage
from ...wallet.did_info import DIDInfo
from ...wallet.did_method import SOV
from ...wallet.key_type import ED25519
from .. import outofband as test_module


class TestOutOfBand(TestCase):
    test_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
    test_did_info = DIDInfo(test_did, test_verkey, None, method=SOV, key_type=ED25519)

    def test_serialize_oob(self):
        invi = InvitationMessage(label="ma sœur", services=[TestOutOfBand.test_did])

        result = test_module.serialize_outofband(
            invi, TestOutOfBand.test_did_info, "http://1.2.3.4:8081"
        )
        assert "?d_m=" in result



import logging

from ..config.injection_context import InjectionContext
from ..config.provider import ClassProvider
from .registry import AnonCredsRegistry

LOGGER = logging.getLogger(__name__)


async def setup(context: InjectionContext) -> None:
    """Set up default resolvers."""
    registry = context.inject_or(AnonCredsRegistry)
    if not registry:
        LOGGER.error("No AnonCredsRegistry instance found in context!!!")
        return

    web_registry = ClassProvider(
        "acapy_agent.anoncreds.default.did_web.registry.DIDWebRegistry",
        # supported_identifiers=[],
        # method_name="did:web",
    ).provide(context.settings, context.injector)
    await web_registry.setup(context)
    registry.register(web_registry)

    legacy_indy_registry = ClassProvider(
        "acapy_agent.anoncreds.default.legacy_indy.registry.LegacyIndyRegistry",
        # supported_identifiers=[],
        # method_name="",
    ).provide(context.settings, context.injector)
    await legacy_indy_registry.setup(context)
    registry.register(legacy_indy_registry)

    # TODO: add context.settings



"""Connection handling admin routes."""

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..cache.base import BaseCache
from ..connections.models.conn_record import ConnRecord, ConnRecordSchema
from ..messaging.models.base import BaseModelError
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.models.paginated_query import (
    PaginatedQuerySchema,
    get_paginated_query_params,
)
from ..messaging.valid import (
    ENDPOINT_EXAMPLE,
    ENDPOINT_VALIDATE,
    GENERIC_DID_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
    UUID4_EXAMPLE,
)
from ..storage.error import StorageError, StorageNotFoundError
from ..wallet.error import WalletError
from .base_manager import BaseConnectionManager


class ConnectionModuleResponseSchema(OpenAPISchema):
    """Response schema for connection module."""


class ConnectionListSchema(OpenAPISchema):
    """Result schema for connection list."""

    results = fields.List(
        fields.Nested(ConnRecordSchema()),
        required=True,
        metadata={"description": "List of connection records"},
    )


class ConnectionMetadataSchema(OpenAPISchema):
    """Result schema for connection metadata."""

    results = fields.Dict(
        metadata={"description": "Dictionary of metadata associated with connection."}
    )


class ConnectionMetadataSetRequestSchema(OpenAPISchema):
    """Request Schema for set metadata."""

    metadata = fields.Dict(
        required=True,
        metadata={"description": "Dictionary of metadata to set for connection."},
    )


class ConnectionMetadataQuerySchema(OpenAPISchema):
    """Query schema for metadata."""

    key = fields.Str(required=False, metadata={"description": "Key to retrieve."})


class ConnectionStaticRequestSchema(OpenAPISchema):
    """Request schema for a new static connection."""

    my_seed = fields.Str(
        required=False, metadata={"description": "Seed to use for the local DID"}
    )
    my_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Local DID", "example": INDY_DID_EXAMPLE},
    )
    their_seed = fields.Str(
        required=False, metadata={"description": "Seed to use for the remote DID"}
    )
    their_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Remote DID", "example": INDY_DID_EXAMPLE},
    )
    their_verkey = fields.Str(
        required=False, metadata={"description": "Remote verification key"}
    )
    their_endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={
            "description": "URL endpoint for other party",
            "example": ENDPOINT_EXAMPLE,
        },
    )
    their_label = fields.Str(
        required=False,
        metadata={"description": "Other party's label for this connection"},
    )
    alias = fields.Str(
        required=False, metadata={"description": "Alias to assign to this connection"}
    )


class ConnectionStaticResultSchema(OpenAPISchema):
    """Result schema for new static connection."""

    my_did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Local DID", "example": INDY_DID_EXAMPLE},
    )
    my_verkey = fields.Str(
        required=True,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "My verification key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    my_endpoint = fields.Str(
        required=True,
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "My URL endpoint", "example": ENDPOINT_EXAMPLE},
    )
    their_did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Remote DID", "example": INDY_DID_EXAMPLE},
    )
    their_verkey = fields.Str(
        required=True,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Remote verification key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    record = fields.Nested(ConnRecordSchema(), required=True)


class ConnectionsListQueryStringSchema(PaginatedQuerySchema):
    """Parameters and validators for connections list request query string."""

    alias = fields.Str(
        required=False, metadata={"description": "Alias", "example": "Barry"}
    )
    invitation_key = fields.Str(
        required=False,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "invitation key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    my_did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "My DID", "example": INDY_DID_EXAMPLE},
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            sorted({label for state in ConnRecord.State for label in state.value})
        ),
        metadata={"description": "Connection state"},
    )
    their_did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "Their DID", "example": INDY_DID_EXAMPLE},
    )
    their_public_did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "Their Public DID", "example": INDY_DID_EXAMPLE},
    )
    their_role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [label for role in ConnRecord.Role for label in role.value]
        ),
        metadata={
            "description": "Their role in the connection protocol",
            "example": ConnRecord.Role.REQUESTER.rfc160,
        },
    )
    connection_protocol = fields.Str(
        required=False,
        validate=validate.OneOf(ConnRecord.SUPPORTED_PROTOCOLS),
        metadata={
            "description": "Connection protocol used",
            "example": "connections/1.0",
        },
    )
    invitation_msg_id = fields.Str(
        required=False,
        metadata={
            "description": "Identifier of the associated Invitation Message",
            "example": UUID4_EXAMPLE,
        },
    )


class ConnectionsConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class ConnIdRefIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection and ref ids."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )

    ref_id = fields.Str(
        required=True,
        metadata={
            "description": "Inbound connection identifier",
            "example": UUID4_EXAMPLE,
        },
    )


class EndpointsResultSchema(OpenAPISchema):
    """Result schema for connection endpoints."""

    my_endpoint = fields.Str(
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "My endpoint", "example": ENDPOINT_EXAMPLE},
    )
    their_endpoint = fields.Str(
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "Their endpoint", "example": ENDPOINT_EXAMPLE},
    )


@docs(
    tags=["connection"],
    summary="Query agent-to-agent connections",
)
@querystring_schema(ConnectionsListQueryStringSchema())
@response_schema(ConnectionListSchema(), 200, description="")
@tenant_authentication
async def connections_list(request: web.BaseRequest):
    """Request handler for searching connection records.

    Args:
        request: aiohttp request object

    Returns:
        The connection list response

    """
    context: AdminRequestContext = request["context"]

    tag_filter = {}
    for param_name in (
        "invitation_id",
        "my_did",
        "their_did",
        "request_id",
        "invitation_key",
        "their_public_did",
        "invitation_msg_id",
    ):
        if param_name in request.query and request.query[param_name] != "":
            tag_filter[param_name] = request.query[param_name]

    post_filter = {}
    if request.query.get("alias"):
        post_filter["alias"] = request.query["alias"]
    if request.query.get("state"):
        post_filter["state"] = list(ConnRecord.State.get(request.query["state"]).value)
    if request.query.get("their_role"):
        post_filter["their_role"] = list(
            ConnRecord.Role.get(request.query["their_role"]).value
        )
    if request.query.get("connection_protocol"):
        post_filter["connection_protocol"] = request.query["connection_protocol"]

    limit, offset, order_by, descending = get_paginated_query_params(request)

    profile = context.profile
    try:
        async with profile.session() as session:
            records = await ConnRecord.query(
                session,
                tag_filter,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
                post_filter_positive=post_filter,
                alt=True,
            )
        results = [record.serialize() for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(tags=["connection"], summary="Fetch a single connection record")
@match_info_schema(ConnectionsConnIdMatchInfoSchema())
@response_schema(ConnRecordSchema(), 200, description="")
@tenant_authentication
async def connections_retrieve(request: web.BaseRequest):
    """Request handler for fetching a single connection record.

    Args:
        request: aiohttp request object

    Returns:
        The connection record response

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]

    profile = context.profile
    try:
        async with profile.session() as session:
            record = await ConnRecord.retrieve_by_id(session, connection_id)
        result = record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["connection"], summary="Fetch connection remote endpoint")
@match_info_schema(ConnectionsConnIdMatchInfoSchema())
@response_schema(EndpointsResultSchema(), 200, description="")
@tenant_authentication
async def connections_endpoints(request: web.BaseRequest):
    """Request handler for fetching connection endpoints.

    Args:
        request: aiohttp request object

    Returns:
        The endpoints response

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]

    profile = context.profile
    connection_mgr = BaseConnectionManager(profile)
    try:
        endpoints = await connection_mgr.get_endpoints(connection_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError, WalletError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(dict(zip(("my_endpoint", "their_endpoint"), endpoints)))


@docs(tags=["connection"], summary="Fetch connection metadata")
@match_info_schema(ConnectionsConnIdMatchInfoSchema())
@querystring_schema(ConnectionMetadataQuerySchema())
@response_schema(ConnectionMetadataSchema(), 200, description="")
@tenant_authentication
async def connections_metadata(request: web.BaseRequest):
    """Handle fetching metadata associated with a single connection record."""
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    key = request.query.get("key", None)

    profile = context.profile
    try:
        async with profile.session() as session:
            record = await ConnRecord.retrieve_by_id(session, connection_id)
            if key:
                result = await record.metadata_get(session, key)
            else:
                result = await record.metadata_get_all(session)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": result})


@docs(tags=["connection"], summary="Set connection metadata")
@match_info_schema(ConnectionsConnIdMatchInfoSchema())
@request_schema(ConnectionMetadataSetRequestSchema())
@response_schema(ConnectionMetadataSchema(), 200, description="")
@tenant_authentication
async def connections_metadata_set(request: web.BaseRequest):
    """Handle fetching metadata associated with a single connection record."""
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    body = await request.json() if request.body_exists else {}

    profile = context.profile
    try:
        async with profile.session() as session:
            record = await ConnRecord.retrieve_by_id(session, connection_id)
            for key, value in body.get("metadata", {}).items():
                await record.metadata_set(session, key, value)
            result = await record.metadata_get_all(session)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": result})


@docs(tags=["connection"], summary="Remove an existing connection record")
@match_info_schema(ConnectionsConnIdMatchInfoSchema())
@response_schema(ConnectionModuleResponseSchema, 200, description="")
@tenant_authentication
async def connections_remove(request: web.BaseRequest):
    """Request handler for removing a connection record.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    profile = context.profile

    try:
        async with profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, connection_id)
            await connection.delete_record(session)
            cache = session.inject_or(BaseCache)
            if cache:
                await cache.clear(f"conn_rec_state::{connection_id}")
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


@docs(tags=["connection"], summary="Create a new static connection")
@request_schema(ConnectionStaticRequestSchema())
@response_schema(ConnectionStaticResultSchema(), 200, description="")
@tenant_authentication
async def connections_create_static(request: web.BaseRequest):
    """Request handler for creating a new static connection.

    Args:
        request: aiohttp request object

    Returns:
        The new connection record

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()

    profile = context.profile
    connection_mgr = BaseConnectionManager(profile)
    try:
        (
            my_info,
            their_info,
            connection,
        ) = await connection_mgr.create_static_connection(
            my_seed=body.get("my_seed") or None,
            my_did=body.get("my_did") or None,
            their_seed=body.get("their_seed") or None,
            their_did=body.get("their_did") or None,
            their_verkey=body.get("their_verkey") or None,
            their_endpoint=body.get("their_endpoint") or None,
            their_label=body.get("their_label") or None,
            alias=body.get("alias") or None,
        )
        response = {
            "my_did": my_info.did,
            "my_verkey": my_info.verkey,
            "my_endpoint": context.settings.get("default_endpoint"),
            "their_did": their_info.did,
            "their_verkey": their_info.verkey,
            "record": connection.serialize(),
        }
    except (WalletError, StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(response)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/connections", connections_list, allow_head=False),
            web.get("/connections/{conn_id}", connections_retrieve, allow_head=False),
            web.get(
                "/connections/{conn_id}/metadata",
                connections_metadata,
                allow_head=False,
            ),
            web.post("/connections/{conn_id}/metadata", connections_metadata_set),
            web.get(
                "/connections/{conn_id}/endpoints",
                connections_endpoints,
                allow_head=False,
            ),
            web.post("/connections/create-static", connections_create_static),
            web.delete("/connections/{conn_id}", connections_remove),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "connection",
            "description": "Connection management",
        }
    )



"""Indy non-revocation interval."""

from time import time
from typing import Optional

from marshmallow import EXCLUDE, fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import INT_EPOCH_EXAMPLE, INT_EPOCH_VALIDATE


class IndyNonRevocationInterval(BaseModel):
    """Indy non-revocation interval."""

    class Meta:
        """NonRevocationInterval metadata."""

        schema_class = "IndyNonRevocationIntervalSchema"

    def __init__(self, fro: Optional[int] = None, to: Optional[int] = None, **kwargs):
        """Initialize non-revocation interval.

        Args:
            fro: earliest time of interest
            to: latest time of interest
            kwargs: additional attributes

        """
        super().__init__(**kwargs)
        self.fro = fro
        self.to = to

    def covers(self, timestamp: Optional[int] = None) -> bool:
        """Whether input timestamp (default now) lies within non-revocation interval.

        Args:
            timestamp: time of interest

        Returns:
            whether input time satisfies non-revocation interval

        """
        timestamp = timestamp or int(time())
        return (self.fro or 0) <= timestamp <= (self.to or timestamp)

    def timestamp(self) -> bool:
        """Return a timestamp that the non-revocation interval covers."""
        return self.to or self.fro or int(time())


class IndyNonRevocationIntervalSchema(BaseModelSchema):
    """Schema to allow serialization/deserialization of non-revocation intervals."""

    class Meta:
        """IndyNonRevocationIntervalSchema metadata."""

        model_class = IndyNonRevocationInterval
        unknown = EXCLUDE

    fro = fields.Int(
        required=False,
        data_key="from",
        validate=INT_EPOCH_VALIDATE,
        metadata={
            "description": "Earliest time of interest in non-revocation interval",
            "strict": True,
            "example": INT_EPOCH_EXAMPLE,
        },
    )
    to = fields.Int(
        required=False,
        validate=INT_EPOCH_VALIDATE,
        metadata={
            "description": "Latest time of interest in non-revocation interval",
            "strict": True,
            "example": INT_EPOCH_EXAMPLE,
        },
    )



"""Utilities for dif presentation exchange attachment.

General Flow:
create_vp ->
make_requirement [create a Requirement from SubmissionRequirements and Descriptors] ->
apply_requirement [filter credentials] ->
merge [return applicable credential list and descriptor_map for presentation_submission]
returns VerifiablePresentation
"""

import logging
import re
from datetime import datetime
from typing import Dict, List, Optional, Sequence, Tuple, Union

from dateutil import tz
from dateutil.parser import ParserError
from dateutil.parser import parse as dateutil_parser
from jsonpath_ng import parse
from pyld import jsonld
from pyld.jsonld import JsonLdProcessor
from unflatten import unflatten
from uuid_utils import uuid4

from ....core.error import BaseError
from ....core.profile import Profile
from ....storage.vc_holder.vc_record import VCRecord
from ....vc.ld_proofs import (
    BbsBlsSignature2020,
    BbsBlsSignatureProof2020,
    DocumentLoader,
    EcdsaSecp256r1Signature2019,
    Ed25519Signature2018,
    Ed25519Signature2020,
    WalletKeyPair,
)
from ....vc.ld_proofs.constants import (
    EXPANDED_TYPE_CREDENTIALS_CONTEXT_V1_VC_TYPE,
    SECURITY_CONTEXT_BBS_URL,
)
from ....vc.vc_di.prove import create_signed_anoncreds_presentation
from ....vc.vc_ld.prove import create_presentation, derive_credential, sign_presentation
from ....wallet.base import BaseWallet, DIDInfo
from ....wallet.default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    ProofPurposeStr,
)
from ....wallet.error import WalletError, WalletNotFoundError
from ....wallet.key_type import BLS12381G2, ED25519, P256
from .pres_exch import (
    Constraints,
    DIFField,
    Filter,
    InputDescriptorMapping,
    InputDescriptors,
    PresentationDefinition,
    PresentationSubmission,
    Requirement,
    SchemaInputDescriptor,
    SchemasInputDescriptorFilter,
    SubmissionRequirements,
)

PRESENTATION_SUBMISSION_JSONLD_CONTEXT = (
    "https://identity.foundation/presentation-exchange/submission/v1"
)
PRESENTATION_SUBMISSION_JSONLD_TYPE = "PresentationSubmission"
PYTZ_TIMEZONE_PATTERN = re.compile(r"(([a-zA-Z]+)(?:\/)([a-zA-Z]+))")
LIST_INDEX_PATTERN = re.compile(r"\[(\W+)\]|\[(\d+)\]")
LOGGER = logging.getLogger(__name__)


class DIFPresExchError(BaseError):
    """Base class for DIF Presentation Exchange related errors."""


class DIFPresExchHandler:
    """Base Presentation Exchange Handler."""

    ISSUE_SIGNATURE_SUITE_KEY_TYPE_MAPPING = {
        Ed25519Signature2018: ED25519,
        Ed25519Signature2020: ED25519,
        EcdsaSecp256r1Signature2019: P256,
    }

    if BbsBlsSignature2020.BBS_SUPPORTED:
        ISSUE_SIGNATURE_SUITE_KEY_TYPE_MAPPING[BbsBlsSignature2020] = BLS12381G2

    DERIVE_SIGNATURE_SUITE_KEY_TYPE_MAPPING = {
        BbsBlsSignatureProof2020: BLS12381G2,
    }
    PROOF_TYPE_SIGNATURE_SUITE_MAPPING = {
        suite.signature_type: suite
        for suite, key_type in ISSUE_SIGNATURE_SUITE_KEY_TYPE_MAPPING.items()
    }
    DERIVED_PROOF_TYPE_SIGNATURE_SUITE_MAPPING = {
        suite.signature_type: suite
        for suite, key_type in DERIVE_SIGNATURE_SUITE_KEY_TYPE_MAPPING.items()
    }

    def __init__(
        self,
        profile: Profile,
        pres_signing_did: Optional[str] = None,
        proof_type: Optional[str] = None,
        reveal_doc: Optional[dict] = None,
    ):
        """Initialize PresExchange Handler."""
        super().__init__()
        self.profile = profile
        self.pres_signing_did = pres_signing_did
        if not proof_type:
            self.proof_type = Ed25519Signature2018.signature_type
        else:
            self.proof_type = proof_type
        self.is_holder = False
        self.reveal_doc_frame = reveal_doc

    async def _get_issue_suite(
        self,
        *,
        issuer_id: str,
        proof_purpose: Optional[ProofPurposeStr] = None,
    ):
        """Get signature suite for signing presentation."""
        proof_purpose = proof_purpose or "assertionMethod"
        did_info = await self._did_info_for_did(issuer_id)
        vm_id_strategy = self.profile.context.inject(BaseVerificationKeyStrategy)
        verification_method = await vm_id_strategy.get_verification_method_id_for_did(
            issuer_id,
            self.profile,
            proof_type=self.proof_type,
            proof_purpose=proof_purpose,
        )

        # Get signature class based on proof type
        SignatureClass = self.PROOF_TYPE_SIGNATURE_SUITE_MAPPING[self.proof_type]

        # Generically create signature class
        return SignatureClass(
            verification_method=verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=self.ISSUE_SIGNATURE_SUITE_KEY_TYPE_MAPPING[SignatureClass],
                public_key_base58=did_info.verkey if did_info else None,
            ),
        )

    async def _get_derive_suite(self):
        """Get signature suite for deriving credentials."""
        # Get signature class based on proof type
        SignatureClass = self.DERIVED_PROOF_TYPE_SIGNATURE_SUITE_MAPPING[
            "BbsBlsSignatureProof2020"
        ]

        # Generically create signature class
        return SignatureClass(
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=self.DERIVE_SIGNATURE_SUITE_KEY_TYPE_MAPPING[SignatureClass],
            ),
        )

    async def _did_info_for_did(self, did: str) -> DIDInfo:
        """Get the did info for specified did.

        If the did starts with did:sov it will remove the prefix for
        backwards compatibility with not fully qualified did.

        Args:
            did (str): The did to retrieve from the wallet.

        Raises:
            WalletNotFoundError: If the did is not found in the wallet.

        Returns:
            DIDInfo: did information

        """
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)

            # If the did starts with did:sov we need to query without
            if did.startswith("did:sov:"):
                return await wallet.get_local_did(did.replace("did:sov:", ""))

            # All other methods we can just query
            return await wallet.get_local_did(did)

    async def get_sign_key_credential_subject_id(
        self, applicable_creds: Sequence[VCRecord]
    ) -> Tuple[Optional[str], Sequence[dict]]:
        """Get the issuer_id and filtered_creds from enclosed credentials subject_ids."""
        issuer_id = None
        filtered_creds_list = []
        if self.proof_type == BbsBlsSignature2020.signature_type:
            reqd_key_type = BLS12381G2
        elif self.proof_type == EcdsaSecp256r1Signature2019.signature_type:
            reqd_key_type = P256
        else:
            reqd_key_type = ED25519
        for cred in applicable_creds:
            if cred.cred_value.get("proof", {}).get("type") == "DataIntegrityProof":
                filtered_creds_list.append(cred.cred_value)
                issuer_id = cred.issuer_id
            elif cred.subject_ids and len(cred.subject_ids) > 0:
                if not issuer_id:
                    for cred_subject_id in cred.subject_ids:
                        if not cred_subject_id.startswith("urn:"):
                            did_info = await self._did_info_for_did(cred_subject_id)
                            if did_info.key_type == reqd_key_type:
                                issuer_id = cred_subject_id
                                filtered_creds_list.append(cred.cred_value)
                                break
                else:
                    if issuer_id in cred.subject_ids:
                        filtered_creds_list.append(cred.cred_value)
                    else:
                        raise DIFPresExchError(
                            "Applicable credentials have different credentialSubject.id, "
                            "multiple proofs are not supported currently"
                        )
        return (issuer_id, filtered_creds_list)

    async def to_requirement(
        self, sr: SubmissionRequirements, descriptors: Sequence[InputDescriptors]
    ) -> Requirement:
        """Return Requirement.

        Args:
            sr: submission_requirement
            descriptors: list of input_descriptors
        Raises:
            DIFPresExchError: If not able to create requirement

        """
        input_descriptors = []
        nested = []
        total_count = 0

        if sr._from:
            if sr._from != "":
                for descriptor in descriptors:
                    if self.contains(descriptor.groups, sr._from):
                        input_descriptors.append(descriptor)
                total_count = len(input_descriptors)
                if total_count == 0:
                    raise DIFPresExchError(f"No descriptors for from: {sr._from}")
        else:
            for submission_requirement in sr.from_nested:
                try:
                    # recursion logic
                    requirement = await self.to_requirement(
                        submission_requirement, descriptors
                    )
                    nested.append(requirement)
                except Exception as err:
                    raise DIFPresExchError(
                        (
                            "Error creating requirement from "
                            f"nested submission_requirements, {err}"
                        )
                    )
            total_count = len(nested)
        count = sr.count
        if sr.rule == "all":
            count = total_count
        requirement = Requirement(
            count=count,
            maximum=sr.maximum,
            minimum=sr.minimum,
            input_descriptors=input_descriptors,
            nested_req=nested,
        )
        return requirement

    async def make_requirement(
        self,
        srs: Sequence[SubmissionRequirements] = None,
        descriptors: Sequence[InputDescriptors] = None,
    ) -> Requirement:
        """Return Requirement.

        Creates and return Requirement with nesting if required
        using to_requirement()

        Args:
            srs: list of submission_requirements
            descriptors: list of input_descriptors
        Raises:
            DIFPresExchError: If not able to create requirement

        """
        if not srs:
            srs = []
        if not descriptors:
            descriptors = []
        if len(srs) == 0:
            requirement = Requirement(
                count=len(descriptors),
                input_descriptors=descriptors,
            )
            return requirement
        requirement = Requirement(
            count=len(srs),
            nested_req=[],
        )
        for submission_requirement in srs:
            try:
                requirement.nested_req.append(
                    await self.to_requirement(submission_requirement, descriptors)
                )
            except Exception as err:
                raise DIFPresExchError(
                    f"Error creating requirement inside to_requirement function, {err}"
                )
        return requirement

    def is_len_applicable(self, req: Requirement, val: int) -> bool:
        """Check and validate requirement minimum, maximum and count.

        Args:
            req: Requirement
            val: int value to check
        Return:
            bool

        """
        if req.count:
            if req.count > 0 and val != req.count:
                return False
        if req.minimum:
            if req.minimum > 0 and req.minimum > val:
                return False
        if req.maximum:
            if req.maximum > 0 and req.maximum < val:
                return False
        return True

    def contains(self, data: Sequence[str], e: str) -> bool:
        """Check for e in data.

        Returns True if e exists in data else return False

        Args:
            data: Sequence of str
            e: str value to check
        Return:
            bool

        """
        data_list = list(data) if data else []
        for k in data_list:
            if e == k:
                return True
        return False

    async def filter_constraints(
        self,
        constraints: Constraints,
        credentials: Sequence[VCRecord],
    ) -> Sequence[VCRecord]:
        """Return list of applicable VCRecords after applying filtering.

        Args:
            constraints: Constraints
            credentials: Sequence of credentials
                to apply filtering on
        Return:
            Sequence of applicable VCRecords

        """
        document_loader = self.profile.inject(DocumentLoader)

        result = []
        for credential in credentials:
            if constraints.subject_issuer == "required" and not self.subject_is_issuer(
                credential=credential
            ):
                continue

            applicable = False
            is_holder_field_ids = self.field_ids_for_is_holder(constraints)
            for field in constraints._fields:
                applicable = await self.filter_by_field(field, credential)
                # all fields in the constraint should be satisfied
                if not applicable:
                    break
                # is_holder with required directive requested for this field
                if applicable and field.id and field.id in is_holder_field_ids:
                    # Missing credentialSubject.id - cannot verify that holder of claim
                    # is same as subject
                    credential_subject_ids = credential.subject_ids
                    if not credential_subject_ids or len(credential_subject_ids) == 0:
                        applicable = False
                        break
                    # Holder of claim is not same as the subject
                    is_holder_same_as_subject = await self.process_constraint_holders(
                        subject_ids=credential_subject_ids
                    )
                    if not is_holder_same_as_subject:
                        applicable = False
                        break
            if not applicable:
                continue
            if constraints.limit_disclosure == "required":
                credential_dict = credential.cred_value
                new_credential_dict = self.reveal_doc(
                    credential_dict=credential_dict, constraints=constraints
                )
                if credential_dict["proof"]["type"] == "DataIntegrityProof":
                    # TODO - don't sign
                    credential = self.create_vcrecord(credential_dict)
                else:
                    derive_suite = await self._get_derive_suite()
                    signed_new_credential_dict = await derive_credential(
                        credential=credential_dict,
                        reveal_document=new_credential_dict,
                        suite=derive_suite,
                        document_loader=document_loader,
                    )
                    credential = self.create_vcrecord(signed_new_credential_dict)
            result.append(credential)
        return result

    def field_ids_for_is_holder(self, constraints: Constraints) -> Sequence[str]:
        """Return list of field ids for whose subject holder verification is requested."""
        reqd_field_ids = set()
        if not constraints.holders:
            reqd_field_ids = []
            return reqd_field_ids
        for holder in constraints.holders:
            if holder.directive == "required" or holder.directive == "preferred":
                reqd_field_ids = set.union(reqd_field_ids, set(holder.field_ids))
        return list(reqd_field_ids)

    async def process_constraint_holders(
        self,
        subject_ids: Sequence[str],
    ) -> bool:
        """Check if holder or subject of claim still controls the identifier."""
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            try:
                for subject_id in subject_ids:
                    await wallet.get_local_did(subject_id.replace("did:sov:", ""))
                self.is_holder = True
                return True
            except (WalletError, WalletNotFoundError):
                return False

    def create_vcrecord(self, cred_dict: dict) -> VCRecord:
        """Return VCRecord from a credential dict."""
        proofs = cred_dict.get("proof") or []
        proof_types = None
        if isinstance(proofs, dict):
            proofs = [proofs]
        if proofs:
            proof_types = [proof.get("type") for proof in proofs]
        contexts = [ctx for ctx in cred_dict.get("@context") if isinstance(ctx, str)]
        if "@graph" in cred_dict:
            for enclosed_data in cred_dict.get("@graph"):
                if (
                    enclosed_data["id"].startswith("urn:")
                    and "credentialSubject" in enclosed_data
                ):
                    cred_dict.update(enclosed_data)
                    del cred_dict["@graph"]
                    break
        given_id = cred_dict.get("id")
        if given_id and self.check_if_cred_id_derived(given_id):
            given_id = str(uuid4())
        # issuer
        issuer = cred_dict.get("issuer")
        if isinstance(issuer, dict):
            issuer = issuer.get("id")

        # subjects
        subject_ids = None
        subjects = cred_dict.get("credentialSubject")
        if subjects:
            if isinstance(subjects, dict):
                subjects = [subjects]
            subject_ids = [subject.get("id") for subject in subjects if ("id" in subject)]
        else:
            cred_dict["credentialSubject"] = {}

        # Schemas
        schemas = cred_dict.get("credentialSchema", [])
        if isinstance(schemas, dict):
            schemas = [schemas]
        schema_ids = [schema.get("id") for schema in schemas]
        document_loader = self.profile.inject(DocumentLoader)
        expanded = jsonld.expand(cred_dict, options={"documentLoader": document_loader})
        types = JsonLdProcessor.get_values(
            expanded[0],
            "@type",
        )
        return VCRecord(
            contexts=contexts,
            expanded_types=types,
            issuer_id=issuer,
            subject_ids=subject_ids,
            proof_types=proof_types,
            given_id=given_id,
            cred_value=cred_dict,
            schema_ids=schema_ids,
        )

    def reveal_doc(self, credential_dict: dict, constraints: Constraints):
        """Generate reveal_doc dict for deriving credential."""
        if not self.reveal_doc_frame:
            derived = {
                "@context": credential_dict.get("@context"),
                "type": credential_dict.get("type"),
                "@explicit": True,
                "@requireAll": True,
                "issuanceDate": {},
                "issuer": {},
            }
            unflatten_dict = {}
            for field in constraints._fields:
                for path in field.paths:
                    jsonpath = parse(path)
                    match = jsonpath.find(credential_dict)
                    if len(match) == 0:
                        continue
                    for match_item in match:
                        full_path = str(match_item.full_path)
                        if bool(LIST_INDEX_PATTERN.search(full_path)):
                            full_path = re.sub(r"\[(\W+)\]|\[(\d+)\]", "[0]", full_path)
                            full_path = full_path.replace(".[", "[")
                        unflatten_dict[full_path] = {}
                        explicit_key_path = None
                        key_list = full_path.split(".")[:-1]
                        for key in key_list:
                            if not explicit_key_path:
                                explicit_key_path = key
                            else:
                                explicit_key_path = explicit_key_path + "." + key
                            unflatten_dict[explicit_key_path + ".@explicit"] = True
                            unflatten_dict[explicit_key_path + ".@requireAll"] = True
            derived = self.new_credential_builder(derived, unflatten_dict)
            # Fix issue related to credentialSubject type property
            if "credentialSubject" in derived.keys():
                if "type" in credential_dict.get("credentialSubject"):
                    derived["credentialSubject"]["type"] = credential_dict[
                        "credentialSubject"
                    ].get("type")
            return derived
        else:
            return self.reveal_doc_frame

    def new_credential_builder(self, new_credential: dict, unflatten_dict: dict) -> dict:
        """Update and return the new_credential.

        Args:
            new_credential: credential dict to be updated and returned
            unflatten_dict: dict with traversal path as key and match_value as value
        Return:
            dict

        """
        new_credential.update(unflatten(unflatten_dict))
        return new_credential

    async def filter_by_field(self, field: DIFField, credential: VCRecord) -> bool:
        """Apply filter on VCRecord.

        Checks if a credential is applicable

        Args:
            field: Field contains filtering spec
            credential: credential to apply filtering on
        Return:
            bool

        """
        credential_dict = credential.cred_value
        for path in field.paths:
            if "$.proof." in path:
                raise DIFPresExchError(
                    "JSON Path expression matching on proof object "
                    "is not currently supported"
                )
            try:
                jsonpath = parse(path)
                match = jsonpath.find(credential_dict)
            except KeyError:
                continue
            if len(match) == 0:
                continue
            for match_item in match:
                # No filter in constraint
                if not field._filter:
                    return True
                if (
                    isinstance(match_item.value, dict)
                    and "type" in match_item.value
                    and "@value" in match_item.value
                ):
                    to_filter_type = match_item.value.get("type")
                    to_filter_value = match_item.value.get("@value")
                    if "integer" in to_filter_type:
                        to_filter_value = int(to_filter_value)
                    elif "dateTime" in to_filter_type or "date" in to_filter_type:
                        to_filter_value = self.string_to_timezone_aware_datetime(
                            to_filter_value
                        )
                    elif "boolean" in to_filter_type:
                        to_filter_value = bool(to_filter_value)
                    elif "double" in to_filter_type or "decimal" in to_filter_type:
                        to_filter_value = float(to_filter_value)
                else:
                    to_filter_value = match_item.value
                if self.validate_patch(to_filter_value, field._filter):
                    return True
        return False

    def string_to_timezone_aware_datetime(self, datetime_str: str) -> datetime:
        """Convert string with PYTZ timezone to datetime for comparison."""
        if PYTZ_TIMEZONE_PATTERN.search(datetime_str):
            result = PYTZ_TIMEZONE_PATTERN.search(datetime_str).group(1)
            datetime_str = datetime_str.replace(result, "")
            return dateutil_parser(datetime_str).replace(tzinfo=tz.gettz(result))
        else:
            return dateutil_parser(datetime_str).replace(tzinfo=tz.UTC)

    def validate_patch(self, to_check: any, _filter: Filter) -> bool:
        """Apply filter on match_value.

        Utility function used in applying filtering to a cred
        by triggering checks according to filter specification

        Args:
            to_check: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        return_val = False
        if _filter._type:
            if self.check_filter_only_type_enforced(_filter):
                if _filter._type == "number":
                    if isinstance(to_check, (int, float)):
                        return True
                elif _filter._type == "string":
                    if isinstance(to_check, str):
                        if _filter.fmt == "date" or _filter.fmt == "date-time":
                            try:
                                to_compare_date = self.string_to_timezone_aware_datetime(
                                    to_check
                                )
                                if isinstance(to_compare_date, datetime):
                                    return True
                            except (ParserError, TypeError):
                                return False
                        else:
                            return True
            else:
                if _filter._type == "number":
                    return_val = self.process_numeric_val(to_check, _filter)
                elif _filter._type == "string":
                    return_val = self.process_string_val(to_check, _filter)
        else:
            if _filter.enums:
                return_val = self.enum_check(val=to_check, _filter=_filter)
            if _filter.const:
                return_val = self.const_check(val=to_check, _filter=_filter)

        if _filter._not:
            return not return_val
        return return_val

    def check_filter_only_type_enforced(self, _filter: Filter) -> bool:
        """Check if only type is specified in filter.

        Args:
            _filter: Filter
        Return:
            bool

        """
        if (
            _filter.pattern is None
            and _filter.minimum is None
            and _filter.maximum is None
            and _filter.min_length is None
            and _filter.max_length is None
            and _filter.exclusive_min is None
            and _filter.exclusive_max is None
            and _filter.const is None
            and _filter.enums is None
        ):
            return True
        else:
            return False

    def process_numeric_val(self, val: any, _filter: Filter) -> bool:
        """Trigger Filter checks.

        Trigger appropriate check for a number type filter,
        according to _filter spec.

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        if _filter.exclusive_max:
            return self.exclusive_maximum_check(val, _filter)
        elif _filter.exclusive_min:
            return self.exclusive_minimum_check(val, _filter)
        elif _filter.minimum:
            return self.minimum_check(val, _filter)
        elif _filter.maximum:
            return self.maximum_check(val, _filter)
        elif _filter.const:
            return self.const_check(val, _filter)
        elif _filter.enums:
            return self.enum_check(val, _filter)
        else:
            return False

    def process_string_val(self, val: any, _filter: Filter) -> bool:
        """Trigger Filter checks.

        Trigger appropriate check for a string type filter,
        according to _filter spec.

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        if _filter.min_length or _filter.max_length:
            return self.length_check(val, _filter)
        elif _filter.pattern:
            return self.pattern_check(val, _filter)
        elif _filter.enums:
            return self.enum_check(val, _filter)
        elif _filter.exclusive_max and _filter.fmt:
            return self.exclusive_maximum_check(val, _filter)
        elif _filter.exclusive_min and _filter.fmt:
            return self.exclusive_minimum_check(val, _filter)
        elif _filter.minimum and _filter.fmt:
            return self.minimum_check(val, _filter)
        elif _filter.maximum and _filter.fmt:
            return self.maximum_check(val, _filter)
        elif _filter.const:
            return self.const_check(val, _filter)
        else:
            return False

    def exclusive_minimum_check(self, val: any, _filter: Filter) -> bool:
        """Exclusiveminimum check.

        Returns True if value greater than filter specified check

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        try:
            if _filter.fmt:
                if _filter.fmt == "date" or _filter.fmt == "date-time":
                    to_compare_date = self.string_to_timezone_aware_datetime(
                        _filter.exclusive_min
                    )
                    given_date = self.string_to_timezone_aware_datetime(str(val))
                    return given_date > to_compare_date
            else:
                try:
                    val = self.is_numeric(val)
                    return val > _filter.exclusive_min
                except DIFPresExchError as err:
                    LOGGER.error(err)
            return False
        except (TypeError, ValueError, ParserError):
            return False

    def exclusive_maximum_check(self, val: any, _filter: Filter) -> bool:
        """Exclusivemaximum check.

        Returns True if value less than filter specified check

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        try:
            if _filter.fmt:
                if _filter.fmt == "date" or _filter.fmt == "date-time":
                    to_compare_date = self.string_to_timezone_aware_datetime(
                        _filter.exclusive_max
                    )
                    given_date = self.string_to_timezone_aware_datetime(str(val))
                    return given_date < to_compare_date
            else:
                try:
                    val = self.is_numeric(val)
                    return val < _filter.exclusive_max
                except DIFPresExchError as err:
                    LOGGER.error(err)
            return False
        except (TypeError, ValueError, ParserError):
            return False

    def maximum_check(self, val: any, _filter: Filter) -> bool:
        """Maximum check.

        Returns True if value less than equal to filter specified check

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        try:
            if _filter.fmt:
                if _filter.fmt == "date" or _filter.fmt == "date-time":
                    to_compare_date = self.string_to_timezone_aware_datetime(
                        _filter.maximum
                    )
                    given_date = self.string_to_timezone_aware_datetime(str(val))
                    return given_date <= to_compare_date
            else:
                try:
                    val = self.is_numeric(val)
                    return val <= _filter.maximum
                except DIFPresExchError as err:
                    LOGGER.error(err)
            return False
        except (TypeError, ValueError, ParserError):
            return False

    def minimum_check(self, val: any, _filter: Filter) -> bool:
        """Minimum check.

        Returns True if value greater than equal to filter specified check

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        try:
            if _filter.fmt:
                if _filter.fmt == "date" or _filter.fmt == "date-time":
                    to_compare_date = self.string_to_timezone_aware_datetime(
                        _filter.minimum
                    )
                    given_date = self.string_to_timezone_aware_datetime(str(val))
                    return given_date >= to_compare_date
            else:
                try:
                    val = self.is_numeric(val)
                    return val >= _filter.minimum
                except DIFPresExchError as err:
                    LOGGER.error(err)
            return False
        except (TypeError, ValueError, ParserError):
            return False

    def length_check(self, val: any, _filter: Filter) -> bool:
        """Length check.

        Returns True if length value string meets the minLength and maxLength specs

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        given_len = len(str(val))
        if _filter.max_length and _filter.min_length:
            if given_len <= _filter.max_length and given_len >= _filter.min_length:
                return True
        elif _filter.max_length and not _filter.min_length:
            if given_len <= _filter.max_length:
                return True
        elif not _filter.max_length and _filter.min_length:
            if given_len >= _filter.min_length:
                return True
        return False

    def pattern_check(self, val: any, _filter: Filter) -> bool:
        """Pattern check.

        Returns True if value string matches the specified pattern

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        if _filter.pattern:
            return bool(re.search(pattern=_filter.pattern, string=str(val)))
        return False

    def const_check(self, val: any, _filter: Filter) -> bool:
        """Const check.

        Returns True if value is equal to filter specified check

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        if val == _filter.const:
            return True
        return False

    def enum_check(self, val: any, _filter: Filter) -> bool:
        """Enum check.

        Returns True if value is contained to filter specified list

        Args:
            val: value to check, extracted from match
            _filter: Filter
        Return:
            bool

        """
        if val in _filter.enums:
            return True
        return False

    def subject_is_issuer(self, credential: VCRecord) -> bool:
        """subject_is_issuer check.

        Returns True if cred issuer_id is in subject_ids

        Args:
            credential: VCRecord
        Return:
            bool

        """
        subject_ids = credential.subject_ids
        for subject_id in subject_ids:
            issuer_id = credential.issuer_id
            if subject_id != "" and subject_id == issuer_id:
                return True
        return False

    async def _credential_match_schema_filter_helper(
        self, credential: VCRecord, filter: Sequence[Sequence[SchemaInputDescriptor]]
    ) -> bool:
        """credential_match_schema for SchemasInputDescriptorFilter.uri_groups."""
        applicable = False
        for uri_group in filter:
            if applicable:
                return True
            for schema in uri_group:
                applicable = self.credential_match_schema(
                    credential=credential, schema_id=schema.uri
                )
                if schema.required and not applicable:
                    break
                if applicable:
                    if schema.uri in [
                        EXPANDED_TYPE_CREDENTIALS_CONTEXT_V1_VC_TYPE,
                    ]:
                        continue
                    else:
                        break
        return applicable

    async def filter_schema(
        self,
        credentials: Sequence[VCRecord],
        schemas: SchemasInputDescriptorFilter,
    ) -> Sequence[VCRecord]:
        """Filter by schema.

        Returns list of credentials where credentialSchema.id or types matched
        with input_descriptors.schema.uri

        Args:
            credentials: list of VCRecords to check
            schemas: list of schemas from the input_descriptors
        Return:
            Sequence of filtered VCRecord

        """
        result = []
        for credential in credentials:
            applicable = False
            applicable = await self._credential_match_schema_filter_helper(
                credential=credential, filter=schemas.uri_groups
            )
            if applicable:
                result.append(credential)
        return result

    def credential_match_schema(self, credential: VCRecord, schema_id: str) -> bool:
        """Credential matching by schema.

        Used by filter_schema to check if credential.schema_ids or credential.types
        matched with schema_id

        Args:
            credential: VCRecord to check
            schema_id: schema uri to check
        Return:
            bool
        """
        if schema_id in credential.schema_ids:
            return True
        if schema_id in credential.expanded_types:
            return True
        return False

    async def filter_creds_record_id(
        self, credentials: Sequence[VCRecord], records_list: Sequence[str]
    ) -> Sequence[VCRecord]:
        """Return filtered list of credentials using records_list."""
        filtered_cred = []
        for credential in credentials:
            if credential.record_id in records_list:
                filtered_cred.append(credential)
        return filtered_cred

    async def apply_requirements(
        self,
        req: Requirement,
        credentials: Sequence[VCRecord],
        records_filter: Optional[dict] = None,
    ) -> dict:
        """Apply Requirement.

        Args:
            req: Requirement
            credentials: Sequence of credentials to check against
            records_filter: dict of input_descriptor ID key to list of credential_json
        Return:
            dict of input_descriptor ID key to list of credential_json
        """
        # Dict for storing descriptor_id keys and list of applicable
        # credentials values
        result = {}
        # Get all input_descriptors attached to the PresentationDefinition
        descriptor_list = req.input_descriptors or []
        for descriptor in descriptor_list:
            # Filter credentials to apply filtering
            # upon by matching each credentialSchema.id
            # or expanded types on each InputDescriptor's schema URIs
            if records_filter and (descriptor.id in records_filter):
                filtered_creds_by_descriptor_id = await self.filter_creds_record_id(
                    credentials, records_filter.get(descriptor.id)
                )
                filtered_by_schema = await self.filter_schema(
                    credentials=filtered_creds_by_descriptor_id,
                    schemas=descriptor.schemas,
                )
            else:
                filtered_by_schema = await self.filter_schema(
                    credentials=credentials,
                    schemas=descriptor.schemas,
                )

            # Filter credentials based upon path expressions specified in constraints
            filtered = await self.filter_constraints(
                constraints=descriptor.constraint,
                credentials=filtered_by_schema,
            )
            if len(filtered) != 0:
                result[descriptor.id] = filtered

        if len(descriptor_list) != 0:
            # Applies min, max or count attributes of submission_requirement
            if self.is_len_applicable(req, len(result)):
                return result
            return {}

        nested_result = []
        cred_uid_descriptors = {}
        # recursion logic for nested requirements
        if req.nested_req:
            for requirement in req.nested_req:
                # recursive call
                result = await self.apply_requirements(
                    requirement, credentials, records_filter
                )
                if result == {}:
                    continue
                # cred_uid_descriptors maps applicable credentials
                # to their respective descriptor.
                # Structure: {cred.given_id: {
                #           desc_id_1: {}
                #      },
                #      ......
                # }
                #  This will be used to construct exclude dict.
                for descriptor_id in result.keys():
                    credential_list = result.get(descriptor_id)
                    for credential in credential_list:
                        cred_id = credential.given_id or credential.record_id
                        if cred_id:
                            cred_uid_descriptors.setdefault(cred_id, {})[
                                descriptor_id
                            ] = {}

                if len(result.keys()) != 0:
                    nested_result.append(result)

        exclude = {}
        for uid in cred_uid_descriptors.keys():
            # Check if number of applicable credentials
            # does not meet requirement specification
            if not self.is_len_applicable(req, len(cred_uid_descriptors[uid])):
                for descriptor_id in cred_uid_descriptors[uid]:
                    # Add to exclude dict
                    # with cred_uid + descriptor_id as key
                    exclude[descriptor_id + uid] = {}
        # merging credentials and excluding credentials that don't satisfy the requirement
        return await self.merge_nested_results(
            nested_result=nested_result, exclude=exclude
        )

    def is_numeric(self, val: any):
        """Check if val is an int or float.

        Args:
            val: to check
        Return:
            numeric value
        Raises:
            DIFPresExchError: Provided value has invalid/incompatible type

        """
        if isinstance(val, float) or isinstance(val, int):
            return val
        elif isinstance(val, str):
            if val.isdigit():
                return int(val)
            else:
                try:
                    return float(val)
                except ValueError:
                    pass
        raise DIFPresExchError("Invalid type provided for comparison/numeric operation.")

    async def merge_nested_results(
        self, nested_result: Sequence[dict], exclude: dict
    ) -> dict:
        """Merge nested results with merged credentials.

        Args:
            nested_result: Sequence of dict containing input_descriptor.id as keys
                and list of creds as values
            exclude: dict containing info about credentials to exclude
        Return:
            dict with input_descriptor.id as keys and merged_credentials_list as values
        """
        result = {}
        for res in nested_result:
            for key in res.keys():
                credentials = res[key]
                uid_dict = {}
                merged_credentials = []

                if key in result:
                    for credential in result[key]:
                        cred_id = credential.given_id or credential.record_id
                        if cred_id and cred_id not in uid_dict:
                            merged_credentials.append(credential)
                            uid_dict[cred_id] = {}

                for credential in credentials:
                    cred_id = credential.given_id or credential.record_id
                    if cred_id and cred_id not in uid_dict:
                        if (key + cred_id) not in exclude:
                            merged_credentials.append(credential)
                            uid_dict[cred_id] = {}
                result[key] = merged_credentials
        return result

    async def create_vp(
        self,
        credentials: Sequence[VCRecord],
        pd: PresentationDefinition,
        challenge: Optional[str] = None,
        domain: Optional[str] = None,
        records_filter: Optional[dict] = None,
        is_holder_override: Optional[bool] = None,
    ) -> Union[Sequence[dict], dict]:
        """Create VerifiablePresentation.

        Args:
            credentials (Sequence[VCRecord]): Sequence of VCRecords.
            pd (PresentationDefinition): PresentationDefinition.
            challenge (str, optional): Challenge string. Defaults to None.
            domain (str, optional): Domain string. Defaults to None.
            records_filter (dict, optional): Records filter dictionary.
            is_holder_override (bool, optional): Override is_holder check.

        Returns:
            Union[Sequence[dict], dict]: VerifiablePresentation.
        """
        document_loader = self.profile.inject(DocumentLoader)
        req = await self.make_requirement(
            srs=pd.submission_requirements, descriptors=pd.input_descriptors
        )

        result = []
        if req.nested_req:
            for nested_req in req.nested_req:
                res = await self.apply_requirements(
                    req=nested_req,
                    credentials=credentials,
                    records_filter=records_filter,
                )
                result.append(res)
        else:
            res = await self.apply_requirements(
                req=req, credentials=credentials, records_filter=records_filter
            )
            result.append(res)

        result_vp = []
        for res in result:
            applicable_creds, descriptor_maps = await self.merge(res)
            applicable_creds_list = []
            for credential in applicable_creds:
                applicable_creds_list.append(credential.cred_value)
            if (
                not self.profile.settings.get("debug.auto_respond_presentation_request")
                and not records_filter
                and len(applicable_creds_list) > 1
            ):
                raise DIFPresExchError(
                    "Multiple credentials are applicable for presentation_definition "
                    f"{pd.id} and --auto-respond-presentation-request setting is not "
                    "enabled. Please specify which credentials should be applied to "
                    "which input_descriptors using record_ids filter."
                )
            submission_property = PresentationSubmission(
                id=str(uuid4()), definition_id=pd.id, descriptor_maps=descriptor_maps
            )
            if self.is_holder or is_holder_override:
                (
                    issuer_id,
                    filtered_creds_list,
                ) = await self.get_sign_key_credential_subject_id(
                    applicable_creds=applicable_creds
                )
                if not issuer_id and len(filtered_creds_list) == 0:
                    vp = await create_presentation(credentials=applicable_creds_list)
                    vp = self.__add_dif_fields_to_vp(vp, submission_property)
                    result_vp.append(vp)
                    continue
                else:
                    applicable_creds_list = filtered_creds_list
                    vp = await create_presentation(credentials=applicable_creds_list)
            else:
                if not self.pres_signing_did:
                    (
                        issuer_id,
                        filtered_creds_list,
                    ) = await self.get_sign_key_credential_subject_id(
                        applicable_creds=applicable_creds
                    )
                    if not issuer_id:
                        vp = await create_presentation(credentials=applicable_creds_list)
                        vp = self.__add_dif_fields_to_vp(vp, submission_property)
                        result_vp.append(vp)
                        continue
                    else:
                        applicable_creds_list = filtered_creds_list
                        vp = await create_presentation(credentials=applicable_creds_list)
                else:
                    issuer_id = self.pres_signing_did
                    vp = await create_presentation(credentials=applicable_creds_list)
            vp["presentation_submission"] = submission_property.serialize()
            if self.proof_type is BbsBlsSignature2020.signature_type:
                vp["@context"].append(SECURITY_CONTEXT_BBS_URL)
            if self.proof_type == ("anoncreds-2023"):
                signed_vp = await create_signed_anoncreds_presentation(
                    profile=self.profile,
                    pres_definition=pd.serialize(),
                    presentation=vp,
                    credentials=applicable_creds_list,
                    challenge=challenge,
                )
            else:
                vp = self.__add_dif_fields_to_vp(vp, submission_property)
                assert issuer_id
                issue_suite = await self._get_issue_suite(
                    issuer_id=issuer_id, proof_purpose="authentication"
                )
                signed_vp = await sign_presentation(
                    presentation=vp,
                    suite=issue_suite,
                    challenge=challenge,
                    document_loader=document_loader,
                )

            result_vp.append(signed_vp)
        if len(result_vp) == 1:
            return result_vp[0]
        return result_vp

    def __add_dif_fields_to_vp(
        self, vp: dict, submission_property: PresentationSubmission
    ) -> dict:
        vp["@context"].append(PRESENTATION_SUBMISSION_JSONLD_CONTEXT)
        vp["type"].append(PRESENTATION_SUBMISSION_JSONLD_TYPE)
        vp["presentation_submission"] = submission_property.serialize()
        if self.proof_type is BbsBlsSignature2020.signature_type:
            vp["@context"].append(SECURITY_CONTEXT_BBS_URL)

        return vp

    def check_if_cred_id_derived(self, id: str) -> bool:
        """Check if credential or credentialSubject id is derived."""
        if id.startswith("urn:bnid:_:c14n"):
            return True
        return False

    async def merge(
        self,
        dict_descriptor_creds: dict,
    ) -> Tuple[Sequence[VCRecord], Sequence[InputDescriptorMapping]]:
        """Return applicable credentials and descriptor_map for attachment.

        Used for generating the presentation_submission property with the
        descriptor_map, maintaining the order in which applicable credential
        list is returned.

        Args:
            dict_descriptor_creds: dict with input_descriptor.id as keys
            and merged_credentials_list
        Return:
            Tuple of applicable credential list and descriptor map
        """
        dict_of_creds = {}
        dict_of_descriptors = {}
        result = []
        descriptors = []
        sorted_desc_keys = sorted(dict_descriptor_creds.keys())
        for desc_id in sorted_desc_keys:
            credentials = dict_descriptor_creds.get(desc_id)
            for cred in credentials:
                cred_id = cred.given_id or cred.record_id
                if cred_id:
                    if cred_id not in dict_of_creds:
                        result.append(cred)
                        dict_of_creds[cred_id] = len(descriptors)
                    if f"{cred_id}-{cred_id}" not in dict_of_descriptors:
                        descriptor_map = InputDescriptorMapping(
                            id=desc_id,
                            fmt="ldp_vc",
                            path=(f"$.verifiableCredential[{dict_of_creds[cred_id]}]"),
                        )
                        descriptors.append(descriptor_map)

        descriptors = sorted(descriptors, key=lambda i: i.id)
        return (result, descriptors)

    async def verify_received_pres(
        self,
        pd: PresentationDefinition,
        pres: Union[Sequence[dict], dict],
    ):
        """Verify credentials received in presentation.

        Args:
            pres: received VerifiablePresentation
            pd: PresentationDefinition
        """
        input_descriptors = pd.input_descriptors
        if isinstance(pres, Sequence):
            for pr in pres:
                descriptor_map_list = pr["presentation_submission"].get("descriptor_map")
                await self.__verify_desc_map_list(
                    descriptor_map_list, pr, input_descriptors
                )
        else:
            descriptor_map_list = pres["presentation_submission"].get("descriptor_map")
            await self.__verify_desc_map_list(
                descriptor_map_list, pres, input_descriptors
            )

    async def __verify_desc_map_list(self, descriptor_map_list, pres, input_descriptors):
        inp_desc_id_constraint_map = {}
        inp_desc_id_schema_one_of_filter = set()
        inp_desc_id_schemas_map = {}
        for input_descriptor in input_descriptors:
            inp_desc_id_constraint_map[input_descriptor.id] = input_descriptor.constraint
            inp_desc_id_schemas_map[input_descriptor.id] = input_descriptor.schemas
            if input_descriptor.schemas.oneof_filter:
                inp_desc_id_schema_one_of_filter.add(input_descriptor.id)
        for desc_map_item in descriptor_map_list:
            desc_map_item_id = desc_map_item.get("id")
            constraint = inp_desc_id_constraint_map.get(desc_map_item_id)
            schema_filter = inp_desc_id_schemas_map.get(desc_map_item_id)
            desc_map_item_path = desc_map_item.get("path")
            jsonpath = parse(desc_map_item_path)
            match = jsonpath.find(pres)
            if len(match) == 0:
                raise DIFPresExchError(
                    f"{desc_map_item_path} path in descriptor_map not applicable"
                )
            for match_item in match:
                try:
                    await self.apply_constraint_received_cred(
                        constraint, match_item.value
                    )
                except BaseError as err:
                    raise DIFPresExchError(
                        f"Constraint specified for {desc_map_item_id} does not "
                        f"apply to the enclosed credential in {desc_map_item_path}"
                        f" Reason: {err.message}"
                    )
                if (
                    not len(
                        await self.filter_schema(
                            credentials=[
                                self.create_vcrecord(cred_dict=match_item.value)
                            ],
                            schemas=schema_filter,
                        )
                    )
                    == 1
                ):
                    raise DIFPresExchError(
                        f"Schema filtering specified in {desc_map_item_id} does not "
                        f"match with the enclosed credential in {desc_map_item_path}"
                    )

    async def restrict_field_paths_one_of_filter(
        self, field_paths: Sequence[str], cred_dict: dict
    ) -> Sequence[str]:
        """Return field_paths that are applicable to oneof_filter."""
        applied_field_paths = []
        for path in field_paths:
            jsonpath = parse(path)
            match = jsonpath.find(cred_dict)
            if len(match) > 0:
                applied_field_paths.append(path)
        return applied_field_paths

    async def apply_constraint_received_cred(
        self, constraint: Constraints, cred_dict: dict
    ) -> None:
        """Evaluate constraint from the request against received credential."""
        """If successful, returns None, else raises an error with reason for failure"""
        fields = constraint._fields
        field_paths = []
        credential = self.create_vcrecord(cred_dict)
        is_limit_disclosure = constraint.limit_disclosure == "required"
        for field in fields:
            if is_limit_disclosure:
                field = await self.get_updated_field(field, cred_dict)
            if not await self.filter_by_field(field, credential):
                raise DIFPresExchError(
                    "Credential is not applicable for field "
                    f"{field.id} with paths {field.paths}"
                )
            field_paths = field_paths + (
                await self.restrict_field_paths_one_of_filter(
                    field_paths=field.paths, cred_dict=cred_dict
                )
            )
        # Selective Disclosure check
        if is_limit_disclosure:
            field_paths = {path.replace("$.", "") for path in field_paths}
            mandatory_paths = {
                "@context",
                "type",
                "issuanceDate",
                "issuer",
                "proof",
                "credentialSubject",
                "id",
            }
            to_remove_from_field_paths = set()
            nested_field_paths = {"credentialSubject": {"id", "type"}}
            for field_path in field_paths:
                if field_path.count(".") >= 1:
                    split_field_path = field_path.split(".")
                    key = ".".join(split_field_path[:-1])
                    value = split_field_path[-1]
                    nested_field_paths = self.build_nested_paths_dict(
                        key, value, nested_field_paths, cred_dict
                    )
                    to_remove_from_field_paths.add(field_path)
            for to_remove_path in to_remove_from_field_paths:
                field_paths.remove(to_remove_path)

            field_paths = set.union(mandatory_paths, field_paths)

            for attrs in cred_dict.keys():
                if attrs not in field_paths:
                    raise DIFPresExchError(
                        "No field in constraints for "
                        f"{'at least one of ' if isinstance(attrs, list) else ''}{attrs}"
                    )
            for nested_attr_key in nested_field_paths:
                nested_attr_values = nested_field_paths[nested_attr_key]
                extracted = self.nested_get(cred_dict, nested_attr_key)
                try:
                    if isinstance(extracted, dict):
                        self.check_attr_in_extracted_dict(extracted, nested_attr_values)
                    elif isinstance(extracted, list):
                        for extracted_dict in extracted:
                            self.check_attr_in_extracted_dict(
                                extracted_dict, nested_attr_values
                            )
                except BaseError as err:
                    raise DIFPresExchError(
                        f"{err.message} under parent path '$.{nested_attr_key}'"
                    )
        return None

    def check_attr_in_extracted_dict(
        self, extracted_dict: dict, nested_attr_values: dict
    ) -> None:
        """Check if keys of extracted_dict exists in nested_attr_values."""
        """If successful, returns None, else raises an error with reason for failure"""
        for attrs in extracted_dict.keys():
            if attrs not in nested_attr_values:
                raise DIFPresExchError(
                    "No field in constraints for "
                    f"{'at least one of ' if isinstance(attrs, list) else ''}{attrs}"
                )
        return None

    def get_dict_keys_from_path(self, derived_cred_dict: dict, path: str) -> List:
        """Return additional_attrs to build nested_field_paths."""
        additional_attrs = []
        mandatory_paths = ["@id", "@type"]
        match_item = self.nested_get(derived_cred_dict, path)
        if isinstance(match_item, dict):
            for key in match_item.keys():
                if key in mandatory_paths:
                    additional_attrs.append(key)
        elif isinstance(match_item, list):
            for key in match_item[0].keys():
                if key in mandatory_paths:
                    additional_attrs.append(key)
        return additional_attrs

    async def get_updated_field(self, field: DIFField, cred: dict) -> DIFField:
        """Return field with updated json path, if necessary."""
        new_paths = []
        for path in field.paths:
            new_paths.append(await self.get_updated_path(cred, path))
        field.paths = new_paths
        return field

    async def get_updated_path(self, cred_dict: dict, json_path: str) -> str:
        """Return updated json path, if necessary."""

        def update_path_recursive_call(path: str, level: int = 0):
            path_split_array = path.split(".", level)
            to_check = path_split_array[-1]
            if "." not in to_check:
                return path
            split_by_index = re.split(r"\[(\d+)\]", to_check, 1)
            if len(split_by_index) > 1:
                jsonpath = parse(split_by_index[0])
                match = jsonpath.find(cred_dict)
                if len(match) > 0:
                    if isinstance(match[0].value, dict):
                        new_path = split_by_index[0] + split_by_index[2]
                        return update_path_recursive_call(new_path, level + 1)
            return update_path_recursive_call(path, level + 1)

        return update_path_recursive_call(json_path)

    def nested_get(self, input_dict: dict, path: str) -> Union[Dict, List]:
        """Return dict or list from nested dict given list of nested_key."""
        jsonpath = parse(path)
        match = jsonpath.find(input_dict)
        if len(match) > 1:
            return_list = []
            for match_item in match:
                return_list.append(match_item.value)
            return return_list
        else:
            return match[0].value

    def build_nested_paths_dict(
        self,
        key: str,
        value: str,
        nested_field_paths: dict,
        cred_dict: dict,
    ) -> dict:
        """Build and return nested_field_paths dict."""
        additional_attrs = self.get_dict_keys_from_path(cred_dict, key)
        value = re.sub(r"\[(\W+)\]|\[(\d+)\]", "", value)
        if key in nested_field_paths.keys():
            nested_field_paths[key].add(value)
        else:
            nested_field_paths[key] = {value}
        if len(additional_attrs) > 0:
            for attr in additional_attrs:
                nested_field_paths[key].add(attr)
        split_key = key.split(".")
        if len(split_key) > 1:
            nested_field_paths.update(
                self.build_nested_paths_dict(
                    ".".join(split_key[:-1]),
                    split_key[-1],
                    nested_field_paths,
                    cred_dict,
                )
            )
        return nested_field_paths



from unittest import TestCase, mock

from ......protocols.didcomm_prefix import DIDCommPrefix
from ...message_types import HANGUP, PROTOCOL_PACKAGE
from ...messages.hangup import Hangup


class TestHangup(TestCase):
    def test_init_type(self):
        """Test initializer."""

        obj = Hangup()
        assert obj._type == DIDCommPrefix.qualify_current(HANGUP)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.hangup.HangupSchema.load")
    def test_deserialize(self, mock_hangup_schema_load):
        """Test deserialization."""

        obj = Hangup()
        hangup = Hangup.deserialize(obj)
        mock_hangup_schema_load.assert_called_once_with(obj)
        assert hangup is mock_hangup_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.hangup.HangupSchema.dump")
    def test_serialize(self, mock_hangup_schema_dump):
        """Test serialization."""

        obj = Hangup()
        hangup_dict = obj.serialize()
        mock_hangup_schema_dump.assert_called_once_with(obj)
        assert hangup_dict is mock_hangup_schema_dump.return_value

    def test_serde(self):
        """Test serialization and deserialization."""

        hangup = Hangup.deserialize({})
        assert hangup._type == DIDCommPrefix.qualify_current(HANGUP)

        hangup_dict = hangup.serialize()
        assert "@id" in hangup_dict
        assert "@type" in hangup_dict

    def test_make_model(self):
        """Test making model."""

        obj = Hangup()
        data = obj.serialize()
        model_instance = Hangup.deserialize(data)
        assert isinstance(model_instance, Hangup)



import pytest

from acapy_agent.wallet.did_method import DIDMethod, DIDMethods, HolderDefinedDid
from acapy_agent.wallet.did_parameters_validation import DIDParametersValidation
from acapy_agent.wallet.error import WalletError
from acapy_agent.wallet.key_type import BLS12381G1, ED25519


@pytest.fixture
def did_methods_registry():
    return DIDMethods()


def test_validate_key_type_uses_didmethod_when_validating_key_type(
    did_methods_registry,
):
    # given
    ed_method = DIDMethod("ed-method", [ED25519])
    did_methods_registry.register(ed_method)
    did_validation = DIDParametersValidation(did_methods_registry)

    # when - then
    assert did_validation.validate_key_type(ed_method, ED25519) is None
    with pytest.raises(WalletError):
        did_validation.validate_key_type(ed_method, BLS12381G1)


def test_validate_key_type_raises_exception_when_validating_unknown_did_method(
    did_methods_registry,
):
    # given
    unknown_method = DIDMethod("unknown", [])
    did_validation = DIDParametersValidation(did_methods_registry)

    # when - then
    with pytest.raises(WalletError):
        did_validation.validate_key_type(unknown_method, ED25519)


def test_set_did_raises_error_when_did_is_provided_and_method_doesnt_allow(
    did_methods_registry,
):
    # given
    ed_method = DIDMethod(
        "derived-did", [ED25519], holder_defined_did=HolderDefinedDid.NO
    )
    did_methods_registry.register(ed_method)
    did_validation = DIDParametersValidation(did_methods_registry)

    # when - then
    with pytest.raises(WalletError):
        did_validation.validate_or_derive_did(
            ed_method, ED25519, b"verkey", "did:edward:self-defined"
        )


def test_validate_or_derive_did_raises_error_when_no_did_is_provided_and_method_requires_one(
    did_methods_registry,
):
    # given
    ed_method = DIDMethod(
        "self-defined-did", [ED25519], holder_defined_did=HolderDefinedDid.REQUIRED
    )
    did_methods_registry.register(ed_method)
    did_validation = DIDParametersValidation(did_methods_registry)

    # when - then
    with pytest.raises(WalletError):
        did_validation.validate_or_derive_did(ed_method, ED25519, b"verkey", did=None)


def test_validate_or_derive_did_raises_exception_when_validating_unknown_did_method(
    did_methods_registry,
):
    # given
    unknown_method = DIDMethod("unknown", [])
    did_validation = DIDParametersValidation(did_methods_registry)

    # when - then
    with pytest.raises(WalletError):
        did_validation.validate_or_derive_did(
            unknown_method, ED25519, b"verkey", did=None
        )
